/* Copyright (c) 2011  Eric Milles <eric.milles@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "Debugger.h"
#include "DeviceDiscovery.h"
#include "ServiceDiscovery.h"
#include "Uncopyable.h"
#include <cassert>
#include <exception>
#include <map>
#include <set>
#include <Poco/Mutex.h>
#include <Poco/ScopedLock.h>


using Poco::FastMutex;
using Poco::ScopedLock;


class DeviceDiscoveryImpl
:
	public ServiceDiscovery::BrowseListener,
	public ServiceDiscovery::QueryListener,
	private Uncopyable
{
	friend class DeviceDiscovery;

	DeviceDiscoveryImpl();
	~DeviceDiscoveryImpl();

	void addListener(DeviceDiscovery::Listener&);
	void removeListener(DeviceDiscovery::Listener&);

	std::set<DeviceDiscovery::Listener*> _listeners;

	void serviceFound(DNSServiceRef, std::string, std::string);
	void serviceLost(DNSServiceRef, std::string, std::string);
	void serviceQueried(DNSServiceRef, std::string, uint16_t, uint16_t,
		const void*, uint32_t);

	DeviceInfo::DeviceType determineDeviceType(
		const ServiceDiscovery::TXTRecord&) const;

	struct ServiceInfo
	{
		std::string name;
		std::string type;
	};

	std::map<DNSServiceRef,ServiceInfo> _serviceInfo;

	std::map<std::string,const DeviceInfo> _devices;

	typedef FastMutex Mutex;
	Mutex _mutex;
};


//------------------------------------------------------------------------------


DeviceDiscoveryImpl& DeviceDiscovery::impl()
{
	static DeviceDiscoveryImpl singleton;
	return singleton;
}


void DeviceDiscovery::browseDevices(Listener& listener)
{
	impl().addListener(listener);
}


void DeviceDiscovery::stopBrowsing(Listener& listener)
{
	impl().removeListener(listener);
}


//------------------------------------------------------------------------------


DeviceDiscoveryImpl::DeviceDiscoveryImpl()
{
	try
	{
		if (ServiceDiscovery::isAvailable())
		{
			ServiceDiscovery::start(
				ServiceDiscovery::browseServices("_raop._tcp.", *this));
		}
	}
	catch (...)
	{
	}
}


DeviceDiscoveryImpl::~DeviceDiscoveryImpl()
{
}


void DeviceDiscoveryImpl::addListener(DeviceDiscovery::Listener& listener)
{
	const ScopedLock<Mutex> lock(_mutex);

	assert(_listeners.find(&listener) == _listeners.end());
	_listeners.insert(&listener);

	// inform listener of all known devices
	for (std::map<std::string,const DeviceInfo>::iterator it = _devices.begin();
		it != _devices.end(); ++it)
	{
		listener.deviceFound(it->second);
	}
}


void DeviceDiscoveryImpl::removeListener(DeviceDiscovery::Listener& listener)
{
	const ScopedLock<Mutex> lock(_mutex);

	_listeners.erase(&listener);
}


void DeviceDiscoveryImpl::serviceFound(DNSServiceRef serviceRef,
	const std::string name, const std::string type)
{
	// leave service browse activity running

	// create service info record
	ServiceInfo info;

	info.name = name;
	info.type = type;

	// create service query activity
	serviceRef = ServiceDiscovery::queryService(
		ServiceDiscovery::fullName(name, type), kDNSServiceType_TXT, *this);

	// store service info for later use
	_serviceInfo[serviceRef] = info;

	// start service query activity
	ServiceDiscovery::start(serviceRef);
}


void DeviceDiscoveryImpl::serviceLost(DNSServiceRef /*serviceRef*/,
	const std::string name, const std::string type)
{
	// leave service browse activity running

	const ScopedLock<Mutex> lock(_mutex);

	// retrieve device info from map
	std::map<std::string,const DeviceInfo>::iterator pos =
		_devices.find(name.substr(name.find_first_of('@') + 1));
	if (pos != _devices.end())
	{
		const DeviceInfo device = pos->second;

		// lose track of old device info
		_devices.erase(pos);

		// inform all listeners of lost device
		for (std::set<DeviceDiscovery::Listener*>::iterator it = _listeners.begin();
			it != _listeners.end(); ++it)
		{
			(**it).deviceLost(device);
		}
	}
}


void DeviceDiscoveryImpl::serviceQueried(DNSServiceRef serviceRef,
	const std::string fullName, const uint16_t /*rrtype*/, const uint16_t rdlen,
	const void* const rdata, const uint32_t /*ttl*/)
{
	// stop service query activity
	ServiceDiscovery::stop(serviceRef);

	// retrieve service info record
	const ServiceInfo info(_serviceInfo[serviceRef]);
	_serviceInfo.erase(serviceRef);

	const ServiceDiscovery::TXTRecord txtRecord(rdata, rdlen);

	if (Debugger::isAvailable())
	{
		Debugger::print("Discovered device \"" + info.name
			+ "\" with TXT record: " + txtRecord.str());
	}

	const DeviceInfo device(
		determineDeviceType(txtRecord),
		info.name.substr(info.name.find_first_of('@') + 1),
		std::make_pair(info.name, info.type),
		true);

	const ScopedLock<Mutex> lock(_mutex);

	// keep track of new device info
	_devices.insert(std::make_pair(device.name(), device));

	// inform all listeners of found device
	for (std::set<DeviceDiscovery::Listener*>::iterator it = _listeners.begin();
		it != _listeners.end(); ++it)
	{
		(**it).deviceFound(device);
	}
}


DeviceInfo::DeviceType DeviceDiscoveryImpl::determineDeviceType(
	const ServiceDiscovery::TXTRecord& txtRecord) const
{
	// AirPort Express 6.1 TXT record looks like:
	// txtvers=1 vn=3 pw=false sr=44100 ss=16 ch=2 cn=1 et=1 ek=1 sv=false
	// sm=false
	
	// AirPort Express 6.2 TXT record looks like:
	// txtvers=1 vn=3 pw=false sr=44100 ss=16 ch=2 cn=0,1 et=0,1 ek=1 sv=false
	// sm=false
	
	// AirPort Express 6.3 TXT record looks like:
	// txtvers=1 vn=3 pw=false sr=44100 ss=16 ch=2 cn=0,1 et=0,1 ek=1 sv=false
	// sm=false tp=TCP,UDP

	// AirPort Express 7.3 TXT record looks like:
	// txtvers=1 ch=2 cn=0,1 ek=1 et=0,1 sv=false da=true sr=44100 ss=16
	// pw=false vn=65537 tp=TCP,UDP vs=101.1

	// AirPort Express 7.4 TXT record looks like:
	// txtvers=1 ch=2 cn=0,1 ek=1 et=0,1 sv=false da=true sr=44100 ss=16
	// pw=false vn=65537 tp=TCP,UDP vs=101.9 am=AirPort4,107 fv=74100.25
	// sf=0x5

	// AirPort Express 7.5 TXT record looks like:
	// txtvers=1 ch=2 cn=0,1 ek=1 et=0,1 sv=false da=true sr=44100 ss=16
	// pw=false vn=65537 tp=TCP,UDP vs=102.2 am=AirPort4,107 fv=75200.14
	// sf=0x5

	// AppleTV TXT record looks like:
	// txtvers=1 ch=2 cn=0,1 et=0,2 sv=false da=true sr=44100 ss=16 pw=false
	// vn=65537 tp=TCP,UDP md=0,1,2 vs=101.8 am=AppleTV1,1 sf=0x4

	// check some common properties of all device types
	assert(txtRecord.has("txtvers") && txtRecord.get("txtvers") == "1");
	assert(txtRecord.has("sr") && txtRecord.get("sr") == "44100");
	assert(txtRecord.has("ss") && txtRecord.get("ss") == "16");
	assert(txtRecord.has("ch") && txtRecord.get("ch") == "2");
	assert(txtRecord.has("pw")
		&& (txtRecord.get("pw") == "false" || txtRecord.get("pw") == "true"));

	// determine device type from some uncommon properties
	if (txtRecord.has("tp") && txtRecord.get("tp") == "UDP")
	{
		throw std::exception("unsupported device type");
	}
	if (txtRecord.has("md"))
	{
		// check some common properties of this device type
		assert(txtRecord.has("cn") && txtRecord.get("cn") == "0,1");
		assert(txtRecord.has("et") && txtRecord.get("et") == "0,2");
		assert(txtRecord.has("da") && txtRecord.get("da") == "true");
		assert(txtRecord.has("sv") && txtRecord.get("sv") == "false");
		assert(txtRecord.has("tp") && txtRecord.get("tp") == "TCP,UDP");
		assert(txtRecord.has("vn") && txtRecord.get("vn") == "65537");
		assert(txtRecord.has("vs"));
		assert(txtRecord.has("am"));
		assert(txtRecord.has("sf"));

		throw std::exception("unsupported device type");
	}
	else if (txtRecord.has("ek"))
	{
		// check some common properties of this device type
		assert(txtRecord.get("ek") == "1");
		assert(txtRecord.has("cn")
			&& (txtRecord.get("cn") == "1" || txtRecord.get("cn") == "0,1"));
		assert(txtRecord.has("et")
			&& (txtRecord.get("et") == "1" || txtRecord.get("et") == "0,1"));
		assert(txtRecord.has("vn")
			&& (txtRecord.get("vn") == "3" || txtRecord.get("vn") == "65537"));

		return DeviceInfo::APX;
	}

	throw std::exception("unknown device type");
}
