/* Copyright (c) 2011  Eric Milles <eric.milles@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "ConnectDialog.h"
#include "DeviceManager.h"
#include "MessageDialog.h"
#include "Options.h"
#include "OutputReformatter.h"
#include "PasswordDialog.h"
#include "RAOPDevice.h"
#include "Uncopyable.h"
#include <cassert>
#include <stdexcept>
#include <Poco/Format.h>
#include <Poco/NotificationCenter.h>
#include <Poco/ScopedLock.h>
#include <Poco/Net/StreamSocket.h>


using Poco::NotificationCenter;
using Poco::Observer;
using Poco::ScopedLock;
using Poco::Net::StreamSocket;


DeviceManager::DeviceManager(const double& balance, const double& volume,
	Player& player)
:
	_balance(balance),
	_volume(volume),
	_writer(false),
	_player(player),
	_observer(*this, &DeviceManager::deviceChanged)
{
	RAOPDevice::staticInit();

	NotificationCenter::defaultCenter().addObserver(_observer);
}


DeviceManager::~DeviceManager()
{
	NotificationCenter::defaultCenter().removeObserver(_observer);

	try
	{
		while (!_devices.empty())
		{
			DeviceMap::iterator pos = _devices.begin();
			delete pos->second;
			_devices.erase(pos);
		}
	}
	catch (...)
	{
	}

	RAOPDevice::staticFree();
}


bool DeviceManager::isAnyDeviceOpen() const
{
	// hold reference to active options
	const OptionsPtr options = Options::getOptions();

	for (DeviceInfoSet::const_iterator it = options->devices().begin();
		it != options->devices().end(); ++it)
	{
		if (isDeviceOpen(*it))
		{
			return true;
		}
	}

	return false;
}


void DeviceManager::openDevices()
{
	bool anyDeviceActivated = false;

	// hold reference to active options
	OptionsPtr options = Options::getOptions();

	for (DeviceInfoSet::iterator it = options->devices().begin();
		it != options->devices().end(); ++it)
	{
		if (options->isActivated(it->name()))
		{
			anyDeviceActivated = true;

			openDevice(*it);
		}
	}

	if (!anyDeviceActivated)
	{
		MessageDialog("No remote speakers are selected for output.").doModal();
		throw std::exception();
	}
}


void DeviceManager::closeDevices()
{
	// hold reference to active options
	OptionsPtr options = Options::getOptions();

	for (DeviceInfoSet::iterator it = options->devices().begin();
		it != options->devices().end(); ++it)
	{
		closeDevice(*it);
	}

	assert(!isAnyDeviceOpen());
}


void DeviceManager::setBalance()
{
	// hold reference to active options
	const OptionsPtr options = Options::getOptions();

	for (DeviceInfoSet::const_iterator it = options->devices().begin();
		it != options->devices().end(); ++it)
	{
		setDeviceBalance(*it, _balance);
	}
}


void DeviceManager::setVolume()
{
	// hold reference to active options
	const OptionsPtr options = Options::getOptions();

	for (DeviceInfoSet::const_iterator it = options->devices().begin();
		it != options->devices().end(); ++it)
	{
		setDeviceVolume(*it, _volume);
	}
}


//------------------------------------------------------------------------------


Device* DeviceManager::createDevice(const DeviceInfo& deviceInfo)
{
	std::auto_ptr<Device> device;

	switch (deviceInfo.type())
	{
	case DeviceInfo::APX:
		device.reset(new RAOPDevice);
		break;
	default:
		MessageDialog(Poco::format(
			"Unable to playback to remote speakers \"%s\".\r\n\r\n"
			"Device type is unsupported at this time.",
			deviceInfo.name())
		).doModal();
		throw std::exception();
	}

	_devices.insert(std::make_pair(deviceInfo.name(), device.get()));

	return device.release();
}


void DeviceManager::destroyDevice(const DeviceInfo& deviceInfo)
{
	const ScopedLock<Mutex> lock(_mutex);

	DeviceMap::iterator pos = _devices.find(deviceInfo.name());
	if (pos != _devices.end())
	{
		delete pos->second;
		_devices.erase(pos);
	}
}


bool DeviceManager::isDeviceOpen(const DeviceInfo& deviceInfo) const
{
	bool isOpen = false;

	const ScopedLock<Mutex> lock(_mutex);

	DeviceMap::const_iterator pos = _devices.find(deviceInfo.name());
	if (pos != _devices.end())
	{
		isOpen = pos->second->isOpen();
	}
	return isOpen;
}


void DeviceManager::openDevice(const DeviceInfo& deviceInfo)
{
	try
	{
		const ScopedLock<Mutex> lock(_mutex);

		if (_devices.find(deviceInfo.name()) == _devices.end())
		{
			RAOPDevice* device =
				dynamic_cast<RAOPDevice*>(createDevice(deviceInfo));

			// run dialog box that will asynchronously resolve service name to
			// host and port, resolve host to IP address and connect to address
			// and port
			ConnectDialog connectDialog(deviceInfo);
			int returnCode = connectDialog.doModal();
			if (returnCode)
			{
				MessageDialog(Poco::format(
						"Unable to connect to remote speakers \"%s\".\r\n\r\n"
						"Error code: %d", deviceInfo.name(), returnCode),
					MB_ICONERROR).doModal();
				throw std::exception();
			}
			StreamSocket& socket = connectDialog.socket();

			// initiate session with remote speakers
			returnCode = device->initiate(socket, true);

			// check if remote speakers require a password
			while (returnCode == 401)
			{
				OptionsPtr options = Options::getOptions();

				// check for password in options
				if (options->getPassword(deviceInfo.name()).empty())
				{
					// prompt user for password
					PasswordDialog passwordDialog(deviceInfo.name());
					if (passwordDialog.doModal(_player.window()))
					{
						throw std::exception();
					}

					options->setPassword(deviceInfo.name(),
						passwordDialog.password(),
						passwordDialog.rememberPassword());
				}

				device->setPassword(options->getPassword(deviceInfo.name()));

				// initiate session with remote speakers again
				returnCode = device->initiate(socket, false);

				// check if password was not accepted
				if (returnCode == 401)
				{
					options->clearPassword(deviceInfo.name());
				}

				// repeat until password is accepted or user cancels
			}

			device->close();

			// check for initiation error
			if (returnCode)
			{
				MessageDialog(Poco::format(
					"Unable to initiate session with remote speakers \"%s\".\r\n\r\n"
					"Error code: %d", deviceInfo.name(), returnCode), MB_ICONERROR).doModal();
				throw std::exception();
			}
		}

		RAOPDevice* device = dynamic_cast<RAOPDevice*>(_devices[deviceInfo.name()]);

		if (!device->isOpen())
		{
			if (!isAnyDeviceOpen())
			{
				RAOPDevice::staticReset();
			}

			// run dialog box that will asynchronously resolve service name to
			// host and port, resolve host to IP address and connect to address
			// and port
			ConnectDialog connectDialog(deviceInfo);
			if (connectDialog.doModal())
			{
				throw std::exception();
			}
			StreamSocket& socket = connectDialog.socket();

			AudioJackStatus audioJackStatus = AUDIO_JACK_CONNECTED;

			// negotiate session parameters with remote speakers
			int returnCode = device->negotiate(socket, audioJackStatus);

			// check if remote speakers require a password
			while (returnCode == 401)
			{
				OptionsPtr options = Options::getOptions();

				// check for password in options
				if (options->getPassword(deviceInfo.name()).empty())
				{
					// prompt user for password
					PasswordDialog passwordDialog(deviceInfo.name());
					if (passwordDialog.doModal(_player.window()))
					{
						throw std::exception();
					}

					options->setPassword(deviceInfo.name(),
						passwordDialog.password(),
						passwordDialog.rememberPassword());
				}

				device->setPassword(options->getPassword(deviceInfo.name()));

				// negotiate session parameters with remote speakers again
				returnCode = device->negotiate(socket, audioJackStatus);

				// check if password was not accepted
				if (returnCode == 401)
				{
					options->clearPassword(deviceInfo.name());
				}

				// repeat until password is accepted or user cancels
			}

			// check for negotiation error
			if (returnCode)
			{
				if (returnCode == 453)
				{
					MessageDialog(Poco::format(
						"Remote speakers \"%s\" are in use by another player.",
						deviceInfo.name())).doModal();
				}
				else
				{
					MessageDialog(
						Poco::format(
							"Unable to connect to remote speakers \"%s\".\r\n\r\n"
							"Error code: %d", deviceInfo.name(), returnCode),
						MB_ICONERROR).doModal();
				}

				throw returnCode;
			}
			else if (audioJackStatus == AUDIO_JACK_DISCONNECTED)
			{
				MessageDialog(Poco::format(
					"Audio jack on remote speakers \"%s\" is not connected.",
					deviceInfo.name())).doModal();
			}

			if (Options::getOptions()->getBalanceAndVolumeControl())
			{
				device->setBalance(_balance);
				device->setVolume(_volume);
			}
		}
	}
	catch (...)
	{
		destroyDevice(deviceInfo);

		Options::getOptions()->setActivated(deviceInfo.name(), false);
		NotificationCenter::defaultCenter().postNotification(
			new DeviceNotification(DeviceNotification::DEACTIVATE, deviceInfo));

		throw;
	}
}


void DeviceManager::closeDevice(const DeviceInfo& deviceInfo)
{
	const ScopedLock<Mutex> lock(_mutex);

	DeviceMap::iterator pos = _devices.find(deviceInfo.name());
	if (pos != _devices.end())
	{
		pos->second->close();
	}
}


void DeviceManager::setDeviceBalance(const DeviceInfo& deviceInfo,
	const double balance)
{
	const ScopedLock<Mutex> lock(_mutex);

	DeviceMap::iterator pos = _devices.find(deviceInfo.name());
	if (pos != _devices.end())
	{
		pos->second->setBalance(balance);
	}
}


void DeviceManager::setDeviceVolume(const DeviceInfo& deviceInfo,
	const double volume)
{
	const ScopedLock<Mutex> lock(_mutex);

	DeviceMap::iterator pos = _devices.find(deviceInfo.name());
	if (pos != _devices.end())
	{
		pos->second->setVolume(volume);
	}
}


void DeviceManager::deviceChanged(DeviceNotification* const notification)
{
	try
	{
		switch (notification->changeType())
		{
		case DeviceNotification::ACTIVATE:
			// open device if open for playback to pick it up immediately
			if (_writer)
			{
				openDevice(notification->deviceInfo());
			}
			break;

		case DeviceNotification::DEACTIVATE:
			destroyDevice(notification->deviceInfo());
			break;
		}
	}
	catch (...)
	{
	}
}


/**
 * Determines if specified DACP remote control identifier is associated with an
 * active device.
 *
 * @param remoteControlId DACP remote control identifier
 * @return <code>true</code> if identifier is valid; <code>false</code> if not
 */
bool DeviceManager::isValidRemoteControlId(const unsigned int remoteControlId) const
{
	bool isValid = false;

	// deadlock would occur if mutex was locked here for AirPort Express 6.1.x
	//const ScopedLock<Mutex> lock(_mutex);

	for (DeviceMap::const_iterator it = _devices.begin(); it != _devices.end(); ++it)
	{
		const unsigned int id = it->second->remoteControlId();
		if (remoteControlId != 0 && id != 0 && remoteControlId == id)
		{
			isValid = true;
			break;
		}
	}

	return isValid;
}


//------------------------------------------------------------------------------


class DeviceWriter
:
	public OutputSink,
	private Uncopyable
{
public:
	DeviceWriter(DeviceManager& manager)
	:
		_manager(manager)
	{
		_manager._writer = true;
	}

	~DeviceWriter()
	{
		_manager._writer = false;
	}

	std::time_t latency(const OutputFormat& format) const
	{
		if (!(format == RAOPDevice::audioFormat()))
		{
			throw std::logic_error("format != RAOPDevice::audioFormat()");
		}

		return 0;
	}

	std::size_t buffered() const
	{
		return 0;
	}

	std::size_t canWrite() const
	{
		return RAOPDevice::audioSendLimit();
	}

	void write(const void* buffer, std::size_t bufferSize)
	{
		const void* encodedBuffer;
		std::size_t encodedBufferSize;
		unsigned int written = 0;

		RAOPDevice::encodeAudio(buffer, bufferSize, encodedBuffer, encodedBufferSize);

		const ScopedLock<DeviceManager::Mutex> lock(_manager._mutex);

		for (DeviceManager::DeviceMap::const_iterator it = _manager._devices.begin();
			it != _manager._devices.end(); ++it)
		{
			Device& device = *(it->second);
			if (device.isOpen())
			{
				try
				{
					device.write(encodedBuffer, encodedBufferSize);
					++written;
				}
				catch (...)
				{
					// quietly close device so playback continues on the others
					device.close();
				}
			}
		}

		if (written > 0)
		{
			RAOPDevice::sentAudio(bufferSize);
		}
		else
		{
			// all devices have errored out; alert the caller
			throw std::exception();
		}
	}

	void flush()
	{
		const ScopedLock<DeviceManager::Mutex> lock(_manager._mutex);

		for (DeviceManager::DeviceMap::const_iterator it = _manager._devices.begin();
			it != _manager._devices.end(); ++it)
		{
			Device& device = *(it->second);
			if (device.isOpen())
			{
				device.flush();
			}
		}
	}

	void reset()
	{
		const ScopedLock<DeviceManager::Mutex> lock(_manager._mutex);

		for (DeviceManager::DeviceMap::const_iterator it = _manager._devices.begin();
			it != _manager._devices.end(); ++it)
		{
			Device& device = *(it->second);
			if (device.isOpen())
			{
				device.reset();
			}
		}
	}

private:
	DeviceManager& _manager;
};


std::auto_ptr<OutputSink> DeviceManager::createOutputSinkForDevices(
	const OutputFormat& format)
{
	std::auto_ptr<OutputSink> outputSink(new DeviceWriter(*this));

	if (!(format == RAOPDevice::audioFormat()))
	{
		outputSink.reset(
			new OutputReformatter(format, RAOPDevice::audioFormat(), outputSink));
	}

	return outputSink;
}
