/* Copyright (c) 2011  Eric Milles <eric.milles@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "Debugger.h"
#include "DeviceManager.h"
#include "Options.h"
#include "OutputComponent.h"
#include "OutputNotification.h"
#include "OutputSink.h"
#include "OutputThread.h"
#include "Platform.h"
#include "RemoteControl.h"
#include <cassert>
#include <memory>
#include <Poco/NotificationCenter.h>
#include <Poco/Runnable.h>
#include <Poco/Thread.h>


using Poco::NotificationCenter;
using Poco::Observer;
using Poco::Runnable;
using Poco::Thread;


class OutputComponentImpl
:
	public Runnable,
	private Uncopyable
{
	friend class OutputComponent;

public:
	void bytesOutput(OutputNotification*);

	void run();

private:
	explicit OutputComponentImpl(Player&);
	~OutputComponentImpl();

	void createOutputChain();

	bool _closeGracefully;
	bool _paused;
	double _balance;
	double _volume;
	unsigned _flushCounter;

	OutputFormat _outputFormat;
	std::auto_ptr<OutputSink> _outputSink;
	Observer<OutputComponentImpl,OutputNotification> _outputObserver;
	OutputComponent::ProgressCallback _progressCallback;

	Player& _player;
	DeviceManager _deviceManager;
	std::auto_ptr<RemoteControl> _remoteControl;

	volatile bool _stopThread;
	Thread _thread;
};


//------------------------------------------------------------------------------


OutputComponent::OutputComponent(Player& player)
:
	_impl(new OutputComponentImpl(player))
{
}


OutputComponent::~OutputComponent()
{
	delete _impl;
}


void OutputComponent::open(const OutputFormat& format)
{
	// ensure component is closed
	close();

	_impl->_closeGracefully = false;
	_impl->_paused = false;
	_impl->_flushCounter = 0;
	_impl->_outputFormat = format;

	_impl->_deviceManager.openDevices();
	_impl->createOutputChain();
}


void OutputComponent::close()
{
	if (_impl->_outputSink.get() != NULL
		&& _impl->_deviceManager.isAnyDeviceOpen())
	{
		if (_impl->_closeGracefully)
		{
			_impl->_outputSink->flush();

			// wait for output sink to drain
			while (_impl->_outputSink->buffered() > 0
				&& _impl->_deviceManager.isAnyDeviceOpen())
			{
				Thread::yield();
			}
		}
		else
		{
			_impl->_outputSink->reset();
		}
	}

	// destroy output chain
	_impl->_outputSink.reset();

	_impl->_closeGracefully = false;
	_impl->_paused = true;
	_impl->_flushCounter = 0;
}


void OutputComponent::reset()
{
	if (_impl->_outputSink.get() != NULL)
	{
		_impl->_outputSink->reset();
	}

	_impl->_paused = false;
}


std::time_t OutputComponent::latency() const
{
	return _impl->_outputSink->latency(_impl->_outputFormat);
}


std::size_t OutputComponent::buffered() const
{
	const std::size_t buffered = _impl->_outputSink->buffered();

	if (buffered > 0)
	{
		if (++_impl->_flushCounter > 16)
		{
			// flush after so many attempts to prevent infinite wait for players
			// that are not nice enough to signal the end of their writing phase
			_impl->_outputSink->flush();
			_impl->_flushCounter = 0;
		}
	}
	else
	{
		if (!_impl->_paused)
		{
			// signal close method that it should wait for output sink to finish
			// writing buffered audio data and it should not reset
			_impl->_closeGracefully = true;
		}
	}

	return buffered;
}


std::size_t OutputComponent::canWrite() const
{
	return (_impl->_paused ? 0 : _impl->_outputSink->canWrite());
}


bool OutputComponent::write(const void* const buffer, const std::size_t length)
{
	if (_impl->_paused)
	{
		// indicate inability to write at this time
		return false;
	}

	// reset indicators because player resumed its writing phase
	_impl->_closeGracefully = false;
	_impl->_flushCounter = 0;

	try
	{
		if (!buffer || !length)
		{
			// player writing phase is over, so flush output buffers
			_impl->_outputSink->flush();
		}
		else
		{
			_impl->_outputSink->write(buffer, length);
		}
	}
	catch (...)
	{
		return false;
	}

	return true;
}


bool OutputComponent::setPaused(const bool state)
{
	const bool previous = _impl->_paused;
	_impl->_paused = state;

	if (_impl->_paused)
	{
		if (Options::getOptions()->getResetOnPause())
		{
			// reset so output stops more quickly
			_impl->_outputSink->reset();
		}
	}
	else
	{
		// reopen devices if open for playback because a long pause closes them
		if (_impl->_outputSink.get() != NULL)
		{
			_impl->_deviceManager.openDevices();
		}
	}

	return previous;
}


void OutputComponent::setBalance(const double balance)
{
	assert(balance >= -1.0 && balance <= 1.0);
	_impl->_balance = balance;
}


void OutputComponent::setVolume(const double volume)
{
	assert(volume >= 0.0 && volume <= 1.0);
	_impl->_volume = volume;
}


void OutputComponent::setProgressCallback(ProgressCallback callback)
{
	_impl->_progressCallback = callback;
}


//------------------------------------------------------------------------------


OutputComponentImpl::OutputComponentImpl(Player& player)
:
	_closeGracefully(false),
	_paused(true),
	_balance(0.0),
	_volume(1.0),
	_flushCounter(0),
	_outputFormat(SampleRate(1), SampleSize(1), ChannelCount(1)),
	_outputObserver(*this, &OutputComponentImpl::bytesOutput),
	_player(player),
	_deviceManager(_balance, _volume, _player),
	_stopThread(false)
{
	_thread.start(*this);

	NotificationCenter::defaultCenter().addObserver(_outputObserver);
}


OutputComponentImpl::~OutputComponentImpl()
{
	NotificationCenter::defaultCenter().removeObserver(_outputObserver);

	_stopThread = true;
	_thread.join();
}


void OutputComponentImpl::bytesOutput(OutputNotification* const notification)
{
	if (_progressCallback)
	{
		_progressCallback(notification->getBytes());
	}
}


void OutputComponentImpl::createOutputChain()
{
	_outputSink = _deviceManager.createOutputSinkForDevices(_outputFormat);

	// wrap output sink so component is non-blocking
	_outputSink.reset(new OutputThread(_outputSink));
}


void OutputComponentImpl::run()
{
	Debugger::setThreadName("OutputComponentImpl::run");

	static const long SLEEP_MSEC = 20;

	bool balVolControlEnabled;
	bool remoteControlEnabled;
	unsigned int pauseCount = 0;
	double previousBalance = _balance;
	double previousVolume = _volume;

	while (!_stopThread)
	{
		try
		{
			Thread::sleep(SLEEP_MSEC);

			// read options then release pointer immediately
			{
				const OptionsPtr options = Options::getOptions();
				balVolControlEnabled = options->getBalanceAndVolumeControl();
				remoteControlEnabled = options->getPlayerControl();
			}

			// monitor pause and balance/volume states

			pauseCount = (_paused && _deviceManager.isAnyDeviceOpen()
				? pauseCount + 1 : 0);

			if (_deviceManager.isAnyDeviceOpen())
			{
				if (pauseCount * SLEEP_MSEC >= 10000)
				{
					// close output devices because player has been paused or
					// stopped for at least ten seconds
					_deviceManager.closeDevices();
				}
				else
				{
					if (balVolControlEnabled)
					{
						if (previousBalance != _balance)
						{
							// balance has changed; set balance of output devices
							_deviceManager.setBalance();
						}

						if (previousVolume != _volume)
						{
							// volume has changed; set volume of output devices
							_deviceManager.setVolume();
						}
					}
				}
			}

			previousBalance = _balance;
			previousVolume = _volume;

			// check for mismatch in state of remote control option and pointer
			if (remoteControlEnabled && _remoteControl.get() == NULL)
			{
				// start remote control listener
				_remoteControl.reset(new RemoteControl(_deviceManager, _player));
			}
			else if (!remoteControlEnabled && _remoteControl.get() != NULL)
			{
				// stop remote control listener
				_remoteControl.reset();
			}
		}
		catch (...)
		{
		}
	}
}
