/* Copyright (c) 2011  Eric Milles <eric.milles@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "OutputReformatter.h"
#include "Platform.h"
#include <cassert>
#include <cmath>
#include <cstring>
#include <stdexcept>


static void ccc_mono_to_stereo(const byte_t* const in, byte_t* const out,
	const int sampleCount, const SampleSize& sampleSize)
{
	// work from back to front to support in-place conversion
	for (int s = sampleCount - 1; s >= 0; --s)
	{
		// duplicate channel data byte by byte
		for (int b = sampleSize - 1; b >= 0; --b)
		{
			out[(s * sampleSize * 2) + b] = in[(s * sampleSize) + b];
			out[(s * sampleSize * 2) + sampleSize + b] = in[(s * sampleSize) + b];
		}
	}
}


static void src_any_to_float_array(const byte_t* const in, float* const out,
	const int sampleCount, const SampleSize& sampleSize)
{
	const int signBit = 1 << ((sampleSize * 8) - 1);
	const float scale = 1.0F * signBit;

	for (int s = 0; s < sampleCount; ++s)
	{
		// load sample byte by byte
		int sample = 0;
		for (int b = sampleSize - 1; b >= 0; --b)
		{
			sample = (sample << 8) | in[(s * sampleSize) + b];
		}

		// extend sign bit if sample is negative
		if (sampleSize < sizeof(int) && sample & signBit)
		{
			for (int b = sampleSize; b < sizeof(int); ++b)
			{
				sample |= 0xFF << (b * 8);
			}
		}

		// scale sample into range [-1.0,1.0]
		out[s] = sample / scale;
	}
}


//------------------------------------------------------------------------------


OutputReformatter::OutputReformatter(const OutputFormat& inFormat,
	const OutputFormat& outFormat, std::auto_ptr<OutputSink> outputSink)
:
	_inFormat(inFormat),
	_outFormat(outFormat),
	_reformatRatio(
		static_cast<double>(_outFormat.sampleRate() * _outFormat.sampleSize() * _outFormat.channelCount())
			/
		static_cast<double>(_inFormat.sampleRate() * _inFormat.sampleSize() * _inFormat.channelCount())),
	_outputSink(outputSink),
	_outputBuffer(_outputSink->canWrite()),
	_srcState(NULL)
{
	if (_inFormat.sampleRate() != _outFormat.sampleRate()
		|| _inFormat.sampleSize() != _outFormat.sampleSize())
	{
		// initialize input buffer size to an amount that will fill output
		// buffer after reformatting
		std::size_t inputBufferSize = static_cast<std::size_t>(
			std::ceil(static_cast<double>(_outputBuffer.size()) / _reformatRatio));

		// convert size from bytes to samples (drop any remainder)
		_inputBuffer.resize(inputBufferSize / _inFormat.sampleSize());

		if (_inFormat.sampleRate() != _outFormat.sampleRate())
		{
			// convert size from bytes to samples (drop any remainder)
			_intermediateBuffer.resize(_outputBuffer.size() / _outFormat.sampleSize());

			// initialize sample rate converter
			int error = 0;
			_srcState = src_new(SRC_SINC_MEDIUM_QUALITY, _inFormat.channelCount(), &error);
			if (_srcState == NULL)
			{
				throw std::runtime_error(src_strerror(error));
			}
		}
	}
}


OutputReformatter::~OutputReformatter()
{
	if (_srcState != NULL)
	{
		src_delete(_srcState);
	}
}


std::time_t OutputReformatter::latency(const OutputFormat& format) const
{
	if (!(format == _inFormat))
	{
		throw std::logic_error("format != _inFormat");
	}

	const std::time_t bufferLatency = _outputBuffer.size() * 1000
		/ (format.sampleRate() * format.sampleSize() * format.channelCount());
	return bufferLatency + _outputSink->latency(_outFormat);
}


std::size_t OutputReformatter::buffered() const
{
	return static_cast<std::size_t>(std::floor(
		static_cast<double>(_outputSink->buffered()) * _reformatRatio));
}


std::size_t OutputReformatter::canWrite() const
{
	// check if intermediate buffer(s) will be used for sample rate and/or size
	// conversion before moving data to output buffer
	if (_inputBuffer.size())
	{
		// convert size from samples to bytes
		return _inputBuffer.size() * _inFormat.sampleSize();
	}
	else
	{
		// return an amount that will fill output buffer after reformatting
		return static_cast<std::size_t>(
			std::floor(static_cast<double>(_outputBuffer.size()) / _reformatRatio));
	}
}


void OutputReformatter::write(const void* const buf, const std::size_t len)
{
	assert(len <= canWrite() && len % _inFormat.sampleSize() == 0);

	int inputSampleCount = len / _inFormat.sampleSize();
	int outputSampleCount = inputSampleCount;

	if (_inFormat.sampleRate() != _outFormat.sampleRate()
		|| _inFormat.sampleSize() != _outFormat.sampleSize())
	{
		const float* sampleBuffer;

		// convert samples to floating-point format
		assert(inputSampleCount <= static_cast<int>(_inputBuffer.size()));
		src_any_to_float_array(static_cast<const byte_t*>(buf), &_inputBuffer[0],
			inputSampleCount, _inFormat.sampleSize());

		if (_inFormat.sampleRate() != _outFormat.sampleRate())
		{
			// prepare for sample rate conversion
			SRC_DATA srcData;
			std::memset(&srcData, 0, sizeof(SRC_DATA));
			srcData.data_in = &_inputBuffer[0];
			srcData.data_out = &_intermediateBuffer[0];
			srcData.input_frames = inputSampleCount / _inFormat.channelCount();
			srcData.output_frames = (_outputBuffer.size() / _outFormat.sampleSize())
				/ _inFormat.channelCount();
			srcData.src_ratio = static_cast<double>(_outFormat.sampleRate())
				/ static_cast<double>(_inFormat.sampleRate());

			if (len == 0)
			{
				// indicate there is no more input so sample rate converter will
				// not maintain any carryover state after next call
				srcData.end_of_input = 1;
			}

			assert(_srcState != NULL);

			// resample input data at output sample rate
			int returnCode = src_process(_srcState, &srcData);
			if (returnCode != 0)
			{
				throw std::runtime_error(src_strerror(returnCode));
			}

			outputSampleCount = srcData.output_frames_gen * _inFormat.channelCount();

			assert(srcData.input_frames_used == srcData.input_frames);
			assert(outputSampleCount <= static_cast<int>(_intermediateBuffer.size()));

			sampleBuffer = &_intermediateBuffer[0];
		}
		else
		{
			sampleBuffer = &_inputBuffer[0];
		}

		// convert samples to integer format
		assert(static_cast<std::size_t>(outputSampleCount * _outFormat.sampleSize())
			<= _outputBuffer.size());
// use of src_float_to_short_array() assumes out sample size of 2
assert(_outFormat.sampleSize() == 2);
		src_float_to_short_array(sampleBuffer,
			reinterpret_cast<short*>(&_outputBuffer[0]), outputSampleCount);
	}

	// write may have been called with no input to flush sample rate converter;
	// check if nothing remains to be written
	if (outputSampleCount < 1)
	{
		return;
	}

	if (_inFormat.channelCount() != _outFormat.channelCount())
	{
		const byte_t* sampleBuffer;

		if (_inFormat.sampleRate() != _outFormat.sampleRate()
			|| _inFormat.sampleSize() != _outFormat.sampleSize())
		{
			// if sample rate and/or size conversion was done, output buffer
			// contains the sample data
			sampleBuffer = &_outputBuffer[0];
		}
		else
		{
			// otherwise use input buffer sample data
			sampleBuffer = static_cast<const byte_t*>(buf);
		}

		assert(static_cast<std::size_t>(outputSampleCount * _outFormat.sampleSize()
			* _outFormat.channelCount()) <= _outputBuffer.size());
// use of ccc_mono_to_stereo() assumes in channel count of 1 and out channel count of 2
assert(_inFormat.channelCount() == 1 && _outFormat.channelCount() == 2);
		ccc_mono_to_stereo(sampleBuffer, &_outputBuffer[0], outputSampleCount,
			_outFormat.sampleSize());

		assert(outputSampleCount % _inFormat.channelCount() == 0);
		outputSampleCount /= _inFormat.channelCount();
		outputSampleCount *= _outFormat.channelCount();
	}

	_outputSink->write(
		&_outputBuffer[0], outputSampleCount * _outFormat.sampleSize());
}


void OutputReformatter::flush()
{
	if (_srcState != NULL)
	{
		// flush sample rate converter with an empty write
		write(NULL, 0);
	}

	_outputSink->flush();
}


void OutputReformatter::reset()
{
	if (_srcState != NULL)
	{
		// reset state of sample rate converter
		int returnCode = src_reset(_srcState);
		if (returnCode != 0)
		{
			assert(returnCode == 0);
		}
	}

	_outputSink->reset();
}
