/* Copyright (c) 2011  Eric Milles <eric.milles@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "Debugger.h"
#include "OutputNotification.h"
#include "OutputThread.h"
#include "Platform.h"
#include <algorithm>
#include <cassert>
#include <cstring>
#include <stdexcept>
#include <Poco/AutoPtr.h>
#include <Poco/NotificationCenter.h>
#include <Poco/ScopedLock.h>
#include <Poco/ScopedUnlock.h>


using Poco::AutoPtr;
using Poco::NotificationCenter;
using Poco::ScopedLock;
using Poco::ScopedUnlock;
using Poco::Thread;


static const std::size_t BUFFER_CAPACITY = 128 * 1024;


OutputThread::OutputThread(std::auto_ptr<OutputSink> outputSink)
:
	_buffer(BUFFER_CAPACITY),
	_bufferAvailability(_buffer.size()),
	_bufferReadIndex(0),
	_bufferWriteIndex(0),
	_outputSink(outputSink),
	_flushThread(false),
	_stopThread(false)
{
	_thread.start(*this);
}


OutputThread::~OutputThread()
{
	try
	{
		_stopThread = true;
		_thread.join();
	}
	catch (...)
	{
	}
}


std::time_t OutputThread::latency(const OutputFormat& format) const
{
	const std::time_t bufferLatency = _buffer.size() * 1000
		/ (format.sampleRate() * format.sampleSize() * format.channelCount());
	return bufferLatency + _outputSink->latency(format);
}


std::size_t OutputThread::buffered() const
{
	const ScopedLock<Mutex> lock(_mutex);

	return _buffer.size() - _bufferAvailability;
}


std::size_t OutputThread::canWrite() const
{
	const ScopedLock<Mutex> lock(_mutex);

	return _bufferAvailability;
}


void OutputThread::write(const void* const buf, const std::size_t len)
{
	const ScopedLock<Mutex> lock(_mutex);

	assert(buf && len);

	if (len > _bufferAvailability)
	{
		throw std::out_of_range("len > _bufferAvailability");
	}

	// write data to buffer
	if (_bufferWriteIndex < _bufferReadIndex
		|| len <= _buffer.size() - _bufferWriteIndex)
	{
		std::memcpy(&_buffer[_bufferWriteIndex], buf, len);
	}
	else
	{
		const std::size_t part1 = _buffer.size() - _bufferWriteIndex;
		std::memcpy(&_buffer[_bufferWriteIndex], buf, part1);
		std::memcpy(&_buffer[0], static_cast<const byte_t*>(buf) + part1, len - part1);
	}

	// adjust buffer availability and write index
	_bufferAvailability -= len;
	_bufferWriteIndex = (_bufferWriteIndex + len) % _buffer.size();
}


void OutputThread::flush()
{
	const ScopedLock<Mutex> lock(_mutex);

	_flushThread = true;
}


void OutputThread::reset()
{
	const ScopedLock<Mutex> lock(_mutex);

	_bufferAvailability = _buffer.size();
	_bufferReadIndex = 0;
	_bufferWriteIndex = 0;
	_flushThread = false;
	_outputSink->reset();
}


void OutputThread::run()
{
	Debugger::setThreadName("OutputThread::run");

#ifdef _WIN32
	DWORD osVersion = GetVersion();
	BYTE majorVersion = LOBYTE(LOWORD(osVersion));
	BYTE minorVersion = HIBYTE(LOWORD(osVersion));
	// Windows 7 does not schedule other threads as aggressively, so yielding is
	// required to allow the other threads to be responsive during playback
	const bool yieldRequired =
		((majorVersion == 0x06 && minorVersion >= 0x01) || majorVersion > 0x06);
#else
	const bool yieldRequired = false; // assume fair preemptive thread scheduler
#endif

	std::size_t canRead = 0;
	std::size_t canWrite = 0;
	std::size_t doWrite = 0;
	std::vector<byte_t> temp;
	std::vector<byte_t>::const_pointer ptr;
	AutoPtr<OutputNotification> notification(new OutputNotification);
	NotificationCenter& notificationCenter = NotificationCenter::defaultCenter();

	while (!_stopThread)
	{
		try
		{
			if (yieldRequired || canRead < canWrite)
			{
				Thread::yield();
			}

			const ScopedLock<Mutex> lock(_mutex);

			canRead = _buffer.size() - _bufferAvailability;
			canWrite = _outputSink->canWrite();

			// determine if data should be written to output sink
			if (canRead > 0 && (canRead >= canWrite || _flushThread))
			{
				doWrite = std::min(canRead, canWrite);

				// check if data to be written is contiguous or not
				if (_bufferReadIndex < _bufferWriteIndex
					|| doWrite <= _buffer.size() - _bufferReadIndex)
				{
					ptr = &_buffer[_bufferReadIndex];
				}
				else
				{
					if (temp.size() < doWrite)
					{
						temp.resize(doWrite);
					}

					const std::size_t part1 = _buffer.size() - _bufferReadIndex;
					std::memcpy(&temp[0], &_buffer[_bufferReadIndex], part1);
					std::memcpy(&temp[part1], &_buffer[0], doWrite - part1);

					ptr = &temp[0];
				}

				{
					ScopedUnlock<Mutex> unlock(_mutex);

					// write data to output sink
					_outputSink->write(ptr, doWrite);
				}

				// check for reset before adjusting state and notifying
				// observers because mutex was released momentarially
				if (_bufferAvailability != _buffer.size())
				{
					// adjust buffer availability and read index
					_bufferAvailability += doWrite;
					_bufferReadIndex = (_bufferReadIndex + doWrite) % _buffer.size();

					// notify any observers of successful output
					notification->setBytes(doWrite);
					notificationCenter.postNotification(notification.get());
				}
			}
			else if (_flushThread)
			{
				// pass on a single flush to output sink when buffer is drained
				_flushThread = false;
				_outputSink->flush();
			}
			else
			{
				Thread::sleep(10);
			}
		}
		catch (...)
		{
			MessageBox(NULL, TEXT("An error ocurred with the remote speaker(s)"),
				TEXT("Remote Speakers otuput"), MB_ICONERROR);
			return;
		}
	}
}
