/* Copyright (c) 2011  Eric Milles <eric.milles@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "Debugger.h"
#include "Plugin.h"
#include "RAOPDevice.h"
#include <cassert>
#include <cstdlib>
#include <cstring>
#include <stdexcept>
#include <vector>
#include <openssl/evp.h>
#include <openssl/md5.h>
#include <Poco/Format.h>


using Poco::Net::SocketAddress;
using Poco::Net::StreamSocket;


/** number of samples per second for RAOP audio data */
#define RAOP_SAMPLES_PER_SECOND 44100

/** number of bits per sample for RAOP audio data */
#define RAOP_BITS_PER_SAMPLE 16

/** number of bits per sample for RAOP audio data */
#define RAOP_CHANNEL_COUNT 2

/** maximum number of samples per channel in RAOP audio data packet */
#define RAOP_PACKET_MAX_SAMPLES_PER_CHANNEL 4096

#define ALAC_HEADER_SIZE 3

#define RTP_HEADER_SIZE 16

#define RAOP_MAX_PACKET_SIZE (RTP_HEADER_SIZE + ALAC_HEADER_SIZE + AES_BLOCK_SIZE \
	+ (RAOP_PACKET_MAX_SAMPLES_PER_CHANNEL * (RAOP_BITS_PER_SAMPLE / 8) * RAOP_CHANNEL_COUNT))

/** RTSP response status codes */
#define RTSP_STATUS_CODE_OK           200
#define RTSP_STATUS_CODE_UNAUTHORIZED 401


static const std::string AUDIO_JACK_STATUS_HEADER("Audio-Jack-Status");
static const std::string AUDIO_LATENCY_HEADER("Audio-Latency");
static const std::string CHALLENGE_RESPONSE_HEADER("Apple-Response");
static const std::string SESSION_HEADER("Session");
static const std::string TRANSPORT_HEADER("Transport");
static const std::string WWW_AUTHENTICATE_HEADER("WWW-Authenticate");


//------------------------------------------------------------------------------


/**
 * Performs ALAC encoding on specified value for specified number of bits.  The
 * result of the encoding is stored in specified buffer at specified offset.
 *
 * @param buffer ALAC buffer
 * @param value data to encode
 * @param length data length
 * @param bitOffset pointer to integer used to store bit offset
 * @param byteOffset pointer to integer used to store byte offset
 */
static void alacEncode(void* const buffer, const int value, std::size_t length,
	std::size_t& bitOffset, std::size_t& byteOffset)
{
	static const byte_t masks[8] = { 0x01, 0x03, 0x07, 0x0F, 0x1F, 0x3F, 0x7F, 0xFF };

	std::size_t writeLength;
	byte_t bitsToWrite;

	if (bitOffset != 0 && bitOffset + length > 8)
	{
		writeLength = 8 - bitOffset;
		bitsToWrite = (byte_t) ((value >> (length - writeLength))
			<< (8 - bitOffset - writeLength));
		static_cast<byte_t*>(buffer)[byteOffset] |= bitsToWrite;
		length -= writeLength;
		bitOffset = 0;
		++byteOffset;
	}

	while (length >= 8)
	{
		bitsToWrite = (byte_t) ((value >> (length - 8)) & 0xFF);
		static_cast<byte_t*>(buffer)[byteOffset] |= bitsToWrite;
		length -= 8;
		bitOffset = 0;
		++byteOffset;
	}

	if (length > 0)
	{
		bitsToWrite = (byte_t) ((value & masks[length])
			<< (8 - bitOffset - length));
		static_cast<byte_t*>(buffer)[byteOffset] |= bitsToWrite;
		bitOffset += length;
		if (bitOffset == 8)
		{
			bitOffset = 0;
			++byteOffset;
		}
	}
}


/**
 * Builds ALAC encoded audio data packet from specified raw audio data.
 */
static void alacBuildPacket(void* const packetBuffer, const std::size_t packetSize,
	const void* const buffer, const std::size_t bufferSize)
{
	std::size_t bitOffset = 0;
	std::size_t byteOffset = 0;
	std::size_t b;

	assert(packetBuffer != NULL && packetSize > 0);
	assert(buffer != NULL && bufferSize > 0);
	assert(bufferSize % 2 == 0);

	// channel count; 0 is monaural, 1 is stereo
	alacEncode(packetBuffer, 1, 3, bitOffset, byteOffset);
	// unknown
	alacEncode(packetBuffer, 0, 4, bitOffset, byteOffset);
	// unknown
	alacEncode(packetBuffer, 0, 12, bitOffset, byteOffset);
	// has_size flag
	alacEncode(packetBuffer, 0, 1, bitOffset, byteOffset);
	// uknown
	alacEncode(packetBuffer, 0, 2, bitOffset, byteOffset);
	// no_compression flag
	alacEncode(packetBuffer, 1, 1, bitOffset, byteOffset);

	// encode audio data; note the endian switch
	for (b = 0; b < bufferSize; b += static_cast<std::size_t>(2))
	{
		alacEncode(packetBuffer, static_cast<const byte_t*>(buffer)[b + 1],
			8, bitOffset, byteOffset);
		alacEncode(packetBuffer, static_cast<const byte_t*>(buffer)[b],
			8, bitOffset, byteOffset);
	}
}


/**
 * Fills specified buffer with random bits.
 *
 * @param buffer byte array
 * @param bufferSize byte array size
 */
static void randomFill(void* const buffer, const std::size_t bufferSize)
{
	assert(buffer != NULL && bufferSize > 0);

	for (std::size_t b = 0; b < bufferSize; ++b)
	{
		static_cast<byte_t*>(buffer)[b] = (byte_t) std::rand();
	}
}


//------------------------------------------------------------------------------


RSA* RAOPDevice::_rsaKey;
AES_KEY RAOPDevice::_aesKey;
char RAOPDevice::_encodedKey[384];
byte_t RAOPDevice::_aesIV[16];
char RAOPDevice::_encodedVector[32];
unsigned short RAOPDevice::_rtpPacketSequenceNumber;
unsigned int RAOPDevice::_rtpTime;
unsigned int RAOPDevice::_rtpSynchronizationSourceId;


void RAOPDevice::staticFree()
{
	if (_rsaKey != NULL)
	{
		// finalize RAOP RSA public key
		RSA_free(_rsaKey);
	}
}


void RAOPDevice::staticInit()
{
	// Apple AirPort Express RSA public key modulus
	static const char modulus[345] =
		"59dE8qLieItsH1WgjrcFRKj6eUWqi+bGLOX1HL3U3GhC/j0Qg90u3sG/1CUtwC"
		"5vOYvfDmFI6oSFXi5ELabWJmT2dKHzBJKa3k9ok+8t9ucRqMd6DZHJ2YCCLlDR"
		"KSKv6kDqnw4UwPdpOMXziC/AMj3Z/lUVX1G7WSHCAWKf1zNS1eLvqr+boEjXuB"
		"OitnZ/bDzPHrTOZz0Dew0uowxf/+sG+NCK3eQJVxqcaJ/vEHKIVd2M+5qL71yJ"
		"Q+87X6oV3eaYvt3zWZYD6z5vYTcrtij2VZ9Zmni/UAaHqn9JdsBWLUEpVviYnh"
		"imNVvYFZeCXg/IdTQ+x4IRdiXNv5hEew==";

	// Apple AirPort Express RSA public key exponent
	static const char exponent[5] = "AQAB";

	int nLength;
	byte_t n[258];
	int eLength;
	byte_t e[3];

	// seed random number generator
	std::srand((unsigned int) time(NULL));

	// base64 decode AirPort Express public key modulus and exponent
	nLength = EVP_DecodeBlock((unsigned char*) n, (unsigned char*) modulus,
		(int) strlen(modulus));
	if (nLength <= 0 || nLength > (int) sizeof(n))
	{
		throw std::runtime_error("EVP_DecodeBlock failed");
	}
	while (n[nLength - 1] == 0 && nLength > 0)
	{
		--nLength;
	}

	eLength = EVP_DecodeBlock((unsigned char*) e, (unsigned char*) exponent,
		(int) strlen(exponent));
	if (eLength <= 0 || eLength > (int) sizeof(e))
	{
		throw std::runtime_error("EVP_DecodeBlock failed");
	}
	while (e[eLength - 1] == 0 && eLength > 0)
	{
		--eLength;
	}

	// initialize RAOP RSA public key
	_rsaKey = RSA_new();
	if (_rsaKey == NULL)
	{
		throw std::runtime_error("RSA_new failed");
	}

	// fill public key components
	_rsaKey->n = BN_bin2bn(n, nLength, _rsaKey->n);
	_rsaKey->e = BN_bin2bn(e, eLength, _rsaKey->e);

	// clear private key components
	_rsaKey->d = NULL;
	_rsaKey->p = NULL;
	_rsaKey->q = NULL;
	_rsaKey->dmp1 = NULL;
	_rsaKey->dmq1 = NULL;
	_rsaKey->iqmp = NULL;
}


void RAOPDevice::staticReset()
{
	byte_t key[16];
	byte_t* encryptedKey;
	int encodedKeyLength;
	int encodedVectorLength;

	// generate AES key
	randomFill(key, sizeof(key));
	if (AES_set_encrypt_key(key, sizeof(key) * 8, &_aesKey))
	{
		throw std::runtime_error("AES_set_encrypt_key failed");
	}

	encryptedKey = (byte_t*) std::malloc(RSA_size(_rsaKey));
	if (encryptedKey == NULL)
	{
		throw std::runtime_error("std::malloc failed");
	}

	// RSA encrypt AES key
	const int returnCode = RSA_public_encrypt(sizeof(key), key, encryptedKey,
		_rsaKey, RSA_PKCS1_OAEP_PADDING);
	if (returnCode < 1)
	{
		free(encryptedKey);
		throw std::runtime_error("RSA_public_encrypt failed");
	}

	// base64 encode RSA-encrypted AES key
	encodedKeyLength = EVP_EncodeBlock((unsigned char*) _encodedKey,
		(unsigned char*) encryptedKey, returnCode);
	free(encryptedKey);
	if (encodedKeyLength <= 0 || encodedKeyLength > (int) sizeof(_encodedKey))
	{
		throw std::runtime_error("EVP_EncodeBlock failed");
	}

	// remove padding from base64-encoded, RSA-encrypted AES key
	while (_encodedKey[encodedKeyLength - 1] == '=' && encodedKeyLength > 0)
	{
		--encodedKeyLength;
	}
	assert(encodedKeyLength < (int) sizeof(_encodedKey));
	_encodedKey[encodedKeyLength] = '\0';

	// generate AES initialization vector
	randomFill(_aesIV, sizeof(_aesIV));

	// base64 encode AES initialization vector
	encodedVectorLength = EVP_EncodeBlock((unsigned char*) _encodedVector,
		(unsigned char*) _aesIV, (int) sizeof(_aesIV));
	if (encodedVectorLength <= 0
		|| encodedVectorLength > (int) sizeof(_encodedVector))
	{
		throw std::runtime_error("EVP_EncodeBlock failed");
	}

	// remove padding from base64-encoded AES initialization vector
	while (_encodedVector[encodedVectorLength - 1] == '=' && encodedVectorLength > 0)
	{
		--encodedVectorLength;
	}
	assert(encodedVectorLength < (int) sizeof(_encodedVector));
	_encodedVector[encodedVectorLength] = '\0';

	// generate initial RTP packet sequence number
	randomFill(&_rtpPacketSequenceNumber, sizeof(unsigned short));

	// generate initial RTP time
	randomFill(&_rtpTime, sizeof(unsigned int));

	// generate initial RTP synchronization source identifier
	randomFill(&_rtpSynchronizationSourceId, sizeof(unsigned int));
}


const OutputFormat& RAOPDevice::audioFormat()
{
	static const OutputFormat format((SampleRate(RAOP_SAMPLES_PER_SECOND)),
		SampleSize(RAOP_BITS_PER_SAMPLE / 8), ChannelCount(RAOP_CHANNEL_COUNT));
	return format;
}


std::size_t RAOPDevice::audioSendLimit()
{
	return (RAOP_PACKET_MAX_SAMPLES_PER_CHANNEL * (RAOP_BITS_PER_SAMPLE / 8) * RAOP_CHANNEL_COUNT);
}


void RAOPDevice::encodeAudio(const void* buffer, const std::size_t bufferSize,
	const void*& outBuffer, std::size_t& outBufferSize)
{
	assert(buffer != NULL && bufferSize > 0);

	// packet buffer is static to avoid allocating it for every audio write
	static byte_t packetBuffer[RAOP_MAX_PACKET_SIZE];
	static std::size_t packetBufferSize;
	static unsigned short packetSequenceNumber = 0;

	// only encrypt a packet once because encryption uses block chaining
	if (packetSequenceNumber != _rtpPacketSequenceNumber)
	{
		packetSequenceNumber = _rtpPacketSequenceNumber;

		// calculate packet size
		packetBufferSize = RTP_HEADER_SIZE + ALAC_HEADER_SIZE + bufferSize;

		// pad packet so it is an integral number of AES blocks
		std::size_t pad = packetBufferSize % AES_BLOCK_SIZE;
		if (pad != 0)
		{
			pad = AES_BLOCK_SIZE - pad;
		}
		packetBufferSize += pad;

		assert(packetBufferSize <= RAOP_MAX_PACKET_SIZE);

		// initialize packet buffer
		std::memset(packetBuffer, 0, packetBufferSize);

		byte_t iv[16];

		// fill in packet header

		// 8-bit packet id
		packetBuffer[0] = '$';
		// 8-bit channel id (0)
		packetBuffer[1] = 0x00;
		// 16-bit rest of packet length
		packetBuffer[2] = HIBYTE((WORD) packetBufferSize - 4);
		packetBuffer[3] = LOBYTE((WORD) packetBufferSize - 4);
		// 2-bit packet version (2), 1-bit padding indicator (0),
		// 1-bit extension indicator (0), 4-bit CSRC count (0)
		packetBuffer[4] = 0x80;
		// 1-bit marker (1), 7-bit payload type (96)
		packetBuffer[5] = 0xE0;
		// 16-bit packet sequence number
		packetBuffer[6] = HIBYTE(_rtpPacketSequenceNumber);
		packetBuffer[7] = LOBYTE(_rtpPacketSequenceNumber);
		// 32-bit timestamp
		packetBuffer[8] = HIBYTE(HIWORD(_rtpTime));
		packetBuffer[9] = LOBYTE(HIWORD(_rtpTime));
		packetBuffer[10] = HIBYTE(LOWORD(_rtpTime));
		packetBuffer[11] = LOBYTE(LOWORD(_rtpTime));
		// 32-bit SSRC
		packetBuffer[12] = HIBYTE(HIWORD(_rtpSynchronizationSourceId));
		packetBuffer[13] = LOBYTE(HIWORD(_rtpSynchronizationSourceId));
		packetBuffer[14] = HIBYTE(LOWORD(_rtpSynchronizationSourceId));
		packetBuffer[15] = LOBYTE(LOWORD(_rtpSynchronizationSourceId));

		// fill in packet payload with audio data
		alacBuildPacket(packetBuffer + RTP_HEADER_SIZE,
			packetBufferSize - RTP_HEADER_SIZE, buffer, bufferSize);

		// make copy of initialization vector because it gets modified
		std::memcpy(iv, _aesIV, 16);

		// encrypt packet payload in place
		AES_cbc_encrypt(packetBuffer + RTP_HEADER_SIZE,
			packetBuffer + RTP_HEADER_SIZE,
			(unsigned long) (packetBufferSize - RTP_HEADER_SIZE),
			&_aesKey, iv, AES_ENCRYPT);
	}

	outBuffer = &packetBuffer;
	outBufferSize = packetBufferSize;
}


void RAOPDevice::sentAudio(const std::size_t size)
{
	// increment packet sequence number
	++_rtpPacketSequenceNumber;

	// increment time
	_rtpTime += ((size / RAOP_CHANNEL_COUNT) / (RAOP_BITS_PER_SAMPLE / 8));
}


//------------------------------------------------------------------------------


RAOPDevice::RAOPDevice()
{
	reinit();

	// generate remote control identifier
	randomFill(&_remoteControlId, sizeof(unsigned int));
}


RAOPDevice::~RAOPDevice()
{
	close();
}


bool RAOPDevice::isOpen() const
{
	return (_rtspSocket.get() != NULL);
}


void RAOPDevice::open()
{
}


void RAOPDevice::close()
{
	try
	{
		if (isOpen())
		{
			// close audio data connection
			_rtpSocket.reset();

			// send teardown message to remote speakers
			sendTeardown();
		}
	}
	catch (...)
	{
	}

	reinit();
}


void RAOPDevice::setBalance(const double /*balance*/)
{
}


void RAOPDevice::setVolume(const double volume)
{
	if (isOpen())
	{
		// convert volume from [0.0,1.0] to -144.0 for 0.0 and [-30.0,0.0] for
		// (0.0,1.0] (remote speakers volume range is [-144.0,0.0] but sound at
		// -60.0 is barely audible with my receiver volume cranked)
		char string[32];
		int length = sprintf_s(string, sizeof(string), "%lf",
			(volume == 0.0 ? -144.0 : (volume - 1.0) * 30.0));
		assert(length > 0 && length < sizeof(string));

		sendSetParameter("volume", string);
	}
}


unsigned int RAOPDevice::remoteControlId() const
{
	return _remoteControlId;
}


std::time_t RAOPDevice::latency(const OutputFormat& format) const
{
	assert(format == RAOPDevice::audioFormat());
	return _audioLatency;
}


std::size_t RAOPDevice::buffered() const
{
	return 0;
}


std::size_t RAOPDevice::canWrite() const
{
	return RAOPDevice::audioSendLimit();
}


void RAOPDevice::write(const void* buffer, const std::size_t length)
{
	assert(buffer != NULL && length > 0);
	assert(_rtpSocket.get() != NULL);

	std::size_t bytesSent = 0;
	while (bytesSent < length)
	{
		const int returnCode = _rtpSocket->sendBytes(
			static_cast<const byte_t*>(buffer) + bytesSent, length - bytesSent);
		if (returnCode <= 0)
		{
			throw std::runtime_error(
				"_rtpSocket->sendBytes returned " + returnCode);
		}
		bytesSent += returnCode;
	}
	if (bytesSent != length)
	{
		throw std::runtime_error("bytesSent != length");
	}
}


void RAOPDevice::flush()
{
	assert(isOpen());
}


void RAOPDevice::reset()
{
	assert(isOpen());
	sendFlush();
}


void RAOPDevice::reinit()
{
	_rtspSocket.reset();
	_rtpSocket.reset();
	_teardownRequired = false;
	_audioLatency = 0;
	_messageSequenceNumber = 0;
	_localSessionId = 0;
	_remoteSessionId = 0;
	//_remoteControlId = 0; // allow this to persist for lifetime of device
	_serverPort = 0;
	_authenticationCasing = 0;
	_authenticationMethod.clear();
	_authenticationRealm.clear();
	_authenticationNonce.clear();
	_authenticationPassword.clear();
}


/**
 * Initiates session with remote speakers on specified socket descriptor.
 *
 * @param firstTime <code>true</code> if first initiation for remote speakers
 * @return zero on success; non-zero on failure
 */
int RAOPDevice::initiate(StreamSocket& socket, const bool firstTime)
{
	if (_rtspSocket.get() == NULL)
	{
		_rtspSocket.reset(new StreamSocket(socket));
		_rtspSocket->setKeepAlive(true);
		_rtspSocket->setLinger(true, 1);
		_rtspSocket->setNoDelay(true);
	}

	// send options message to remote speakers
	int returnCode = sendOptions(firstTime);
	if (returnCode != RTSP_STATUS_CODE_OK)
	{
		return returnCode;
	}

	return 0;
}


/**
 * Negotiates session parameters with remote speakers on specified socket.
 *
 * @param audioJackStatus reference used to return status of remote speakers
 *                        audio jack
 * @return zero on success; non-zero on failure
 */
int RAOPDevice::negotiate(StreamSocket& socket, AudioJackStatus& audioJackStatus)
{
	if (_rtspSocket.get() == NULL)
	{
		_rtspSocket.reset(new StreamSocket(socket));
		_rtspSocket->setKeepAlive(true);
		_rtspSocket->setLinger(true, 1);
		_rtspSocket->setNoDelay(true);
	}

	// send announce message to remote speakers
	int returnCode = sendAnnounce();
	if (returnCode != RTSP_STATUS_CODE_OK)
	{
		return returnCode;
	}

	// send setup message to remote speakers
	returnCode = sendSetup(audioJackStatus);
	if (returnCode != RTSP_STATUS_CODE_OK)
	{
		return returnCode;
	}

	// send record message to remote speakers
	returnCode = sendRecord();
	if (returnCode != RTSP_STATUS_CODE_OK)
	{
		return returnCode;
	}

	// open audio data connection
	assert(_rtpSocket.get() == NULL);
	assert(_rtspSocket.get() != NULL);
	assert(_serverPort > 0);

	_rtpSocket.reset(new StreamSocket(
		SocketAddress(_rtspSocket->peerAddress().host(), _serverPort)));
	_rtpSocket->setKeepAlive(true);
	_rtpSocket->setLinger(true, 1);
	_rtpSocket->setNoDelay(true);

	return 0;
}


/**
 * Sets password used by device when authentication is required.
 *
 * @param password remote speakers password string
 */
void RAOPDevice::setPassword(const std::string& password)
{
	_authenticationPassword = password;
}


//------------------------------------------------------------------------------


/**
 * Sends RTSP OPTIONS message
 *
 * @param sendAuthenticityChallenge if <code>true</code>, send cryptographic key
 *                                  challenge to remote speakers
 * @return zero on success; non-zero on failure
 */
int RAOPDevice::sendOptions(const bool sendAuthenticityChallenge)
{
	byte_t challengeNonce[16];
	char requestHead[1024];
	int returnCode;
	struct Response response;

	std::memset(requestHead, 0, sizeof(requestHead));

	if (sendAuthenticityChallenge)
	{
		char encodedChallengeNonce[32];
		int encodedChallengeNonceLength;
		int requestHeadLength;

		// generate random challenge nonce
		randomFill(challengeNonce, sizeof(challengeNonce));

		// base64 encode challenge nonce
		encodedChallengeNonceLength = EVP_EncodeBlock(
			(unsigned char*) encodedChallengeNonce,
			(unsigned char*) challengeNonce, (int) sizeof(challengeNonce));
		assert(encodedChallengeNonceLength > 0
			&& encodedChallengeNonceLength <= (int) sizeof(encodedChallengeNonce));

		// remove padding from base64-encoded challenge nonce
		while (encodedChallengeNonce[encodedChallengeNonceLength - 1] == '='
			&& encodedChallengeNonceLength > 0)
		{
			--encodedChallengeNonceLength;
		}
		assert(encodedChallengeNonceLength < (int) sizeof(encodedChallengeNonce));
		encodedChallengeNonce[encodedChallengeNonceLength] = '\0';

		requestHeadLength = sprintf_s(requestHead, sizeof(requestHead),
			"Apple-Challenge: %s\r\n", encodedChallengeNonce);
		assert(requestHeadLength > 0 && requestHeadLength < sizeof(requestHead));
	}

	sendRequestReceiveResponse("OPTIONS", requestHead, "", response);
	if (response.statusCode != RTSP_STATUS_CODE_OK
		// allow unauthorized conditioin to flow through so challenge is checked
		&& response.statusCode != RTSP_STATUS_CODE_UNAUTHORIZED)
	{
		return response.statusCode;
	}

	if (sendAuthenticityChallenge)
	{
		int length;
		char encodedResponseNonce[384];
		byte_t encryptedResponseNonce[258];
		byte_t* responseNonce;

		if (response.headers.find(CHALLENGE_RESPONSE_HEADER) == response.headers.end())
		{
			return -200000;
		}

		const std::string str = response.headers[CHALLENGE_RESPONSE_HEADER];
		length = (int) str.length();
		strncpy_s(encodedResponseNonce, sizeof(encodedResponseNonce), str.c_str(),
			(size_t) length);

		// pad base64 string as necessary
		while (length % 4 != 0)
		{
			encodedResponseNonce[length++] = '=';
		}
		encodedResponseNonce[length] = '\0';

		// base64 decode RSA-encrypted response nonce
		length = EVP_DecodeBlock((unsigned char*) encryptedResponseNonce,
			(unsigned char*) encodedResponseNonce, length);
		if (length <= 0 || length > (int) sizeof(encryptedResponseNonce))
		{
			return -200001;
		}
		while (encryptedResponseNonce[length - 1] == 0 && length > 0)
		{
			--length;
		}

		responseNonce = (byte_t*) std::malloc(RSA_size(_rsaKey));
		if (responseNonce == NULL)
		{
			return -200002;
		}

		// RSA decrypt response nonce
		returnCode = RSA_public_decrypt(length, encryptedResponseNonce,
			responseNonce, _rsaKey, RSA_PKCS1_PADDING);
		if (returnCode < 1)
		{
			std::free(responseNonce);
			return -200003;
		}
		assert(returnCode >= sizeof(challengeNonce));

		// compare challenge nonce to response nonce; they should match if
		// remote speakers posseses private counterpart to RSA public key used
		if (std::memcmp(challengeNonce, responseNonce, sizeof(challengeNonce)) != 0)
		{
			std::free(responseNonce);
			return -200004;
		}

		std::free(responseNonce);
	}

	return response.statusCode;
}


/**
 * Sends RTSP ANNOUNCE message
 *
 * @return zero on success; non-zero on failure
 */
int RAOPDevice::sendAnnounce()
{
	const unsigned int maximumSamplesPerChannel = RAOP_PACKET_MAX_SAMPLES_PER_CHANNEL;
	const unsigned int bitsPerSample = RAOP_BITS_PER_SAMPLE;
	const unsigned int samplesPerSecond = RAOP_SAMPLES_PER_SECOND;
	const char* encodedKey = _encodedKey;
	const char* encodedIV = _encodedVector;

	char localAddress[16];
	char remoteAddress[16];
	int requestBodyLength;
	char requestBody[1024];
	int requestHeadLength;
	char requestHead[1024];
	struct Response response;

	// generate local session identifier
	randomFill(&_localSessionId, sizeof(unsigned int));

	// get dotted-decimal IP address of local socket endpoint
	strcpy_s(localAddress, sizeof(localAddress),
		_rtspSocket->address().host().toString().c_str());

	// get dotted-decimal IP address of remote socket endpoint
	strcpy_s(remoteAddress, sizeof(remoteAddress),
		_rtspSocket->peerAddress().host().toString().c_str());

	requestBodyLength = sprintf_s(requestBody, sizeof(requestBody),
		"v=0\r\n"
		"o=iTunes %u 0 IN IP4 %s\r\n"
		"s=iTunes\r\n"
		"c=IN IP4 %s\r\n"
		"t=0 0\r\n"
		"m=audio 0 RTP/AVP 96\r\n"
		"a=rtpmap:96 AppleLossless\r\n"
		"a=fmtp:96 %u 0 %u 40 10 14 2 255 0 0 %u\r\n"
		"a=rsaaeskey:%s\r\n"
		"a=aesiv:%s\r\n",
		_localSessionId, localAddress, remoteAddress,
		maximumSamplesPerChannel, bitsPerSample, samplesPerSecond,
		encodedKey, encodedIV);
	assert(requestBodyLength > 0 && requestBodyLength < sizeof(requestBody));

	requestHeadLength = sprintf_s(requestHead, sizeof(requestHead),
		"Content-Type: application/sdp\r\n"
		"Content-Length: %u\r\n",
		requestBodyLength);
	assert(requestHeadLength > 0 && requestHeadLength < sizeof(requestHead));

	sendRequestReceiveResponse("ANNOUNCE", requestHead, requestBody, response);

	return response.statusCode;
}


/**
 * Sends RTSP SETUP message
 *
 * @param audioJackStatus reference used to return status of remote speakers
 *                        audio jack
 * @return zero on success; non-zero on failure
 */
int RAOPDevice::sendSetup(AudioJackStatus& audioJackStatus)
{
	int requestHeadLength;
	char requestHead[1024];
	struct Response response;
	char* string;
	char* token;
	char* state;
	char* value;

	requestHeadLength = sprintf_s(requestHead, sizeof(requestHead),
		"Transport: RTP/AVP/TCP;unicast;interleaved=0-1;mode=record\r\n");
	assert(requestHeadLength > 0 && requestHeadLength < sizeof(requestHead));

	sendRequestReceiveResponse("SETUP", requestHead, "", response);
	if (response.statusCode != RTSP_STATUS_CODE_OK)
	{
		return response.statusCode;
	}

	if (response.headers.find(SESSION_HEADER) == response.headers.end())
	{
		return -200100;
	}

	// parse session header
	sscanf_s(response.headers[SESSION_HEADER].c_str(), "%x", &_remoteSessionId);

	if (response.headers.find(TRANSPORT_HEADER) == response.headers.end())
	{
		return -200101;
	}

	// copy transport header so it can be modified
	string = _strdup(response.headers[TRANSPORT_HEADER].c_str());
	if (string == NULL)
	{
		return -200102;
	}

	// parse transport header
	token = strtok_s(string, ";", &state);
	do
	{
		assert(token != NULL && token[0] != '\0');

		value = strchr(token, '=');
		if (value != NULL)
		{
			// break on the equal sign
			value[0] = '\0';
			value++;

			if (std::strcmp("server_port", token) == 0)
			{
				_serverPort = static_cast<unsigned short>(std::atoi(value));
			}
		}
	}
	while((token = strtok_s(NULL, ";", &state)) != NULL);

	std::free(string);

	if (response.headers.find(AUDIO_JACK_STATUS_HEADER) != response.headers.end())
	{
		// parse audio jack status header
		audioJackStatus =
			(response.headers[AUDIO_JACK_STATUS_HEADER] == "disconnected"
				? AUDIO_JACK_DISCONNECTED : AUDIO_JACK_CONNECTED);
	}

	return response.statusCode;
}


/**
 * Sends RTSP RECORD message
 *
 * @return zero on success; non-zero on failure
 */
int RAOPDevice::sendRecord()
{
	int requestHeadLength;
	char requestHead[1024];
	struct Response response;

	requestHeadLength = sprintf_s(requestHead, sizeof(requestHead),
		"Session: %.8X\r\n"
		"Range: npt=0-\r\n"
		"RTP-Info: seq=%hu;rtptime=%u\r\n",
		_remoteSessionId, _rtpPacketSequenceNumber, _rtpTime);
	assert(requestHeadLength > 0 && requestHeadLength < sizeof(requestHead));

	sendRequestReceiveResponse("RECORD", requestHead, "", response);
	if (response.statusCode != RTSP_STATUS_CODE_OK)
	{
		return response.statusCode;
	}

	if (response.headers.find(AUDIO_LATENCY_HEADER) != response.headers.end())
	{
		// parse audio latency header
		sscanf_s(response.headers[AUDIO_LATENCY_HEADER].c_str(), "%u", &_audioLatency);
	}

	// indicate that teardown is required on disconnect because record message
	// was received successfully and a session is now fully established
	_teardownRequired = true;

	return response.statusCode;
}


/**
 * Sends RTSP TEARDOWN message
 *
 * @return zero on success; non-zero on failure
 */
int RAOPDevice::sendTeardown()
{
	if (!_teardownRequired)
		return 0;

	int requestHeadLength;
	char requestHead[1024];
	struct Response response;

	requestHeadLength = sprintf_s(requestHead, sizeof(requestHead),
		"Session: %.8X\r\n", _remoteSessionId);
	assert(requestHeadLength > 0 && requestHeadLength < sizeof(requestHead));

	sendRequestReceiveResponse("TEARDOWN", requestHead, "", response);

	return response.statusCode;
}


/**
 * Sends RTSP SET_PARAMETER message
 *
 * @param paramaterName parameter name
 * @param parameterValue parameter value
 * @return zero on success; non-zero on failure
 */
int RAOPDevice::sendSetParameter(
	const std::string& parameterName, const std::string& parameterValue)
{
	int requestBodyLength;
	char requestBody[1024];
	int requestHeadLength;
	char requestHead[1024];
	struct Response response;

	assert(!parameterName.empty());
	assert(!parameterValue.empty());

	requestBodyLength = sprintf_s(requestBody, sizeof(requestBody),
		"%s: %s\r\n", parameterName.c_str(), parameterValue.c_str());
	assert(requestBodyLength > 0 && requestBodyLength < sizeof(requestBody));

	requestHeadLength = sprintf_s(requestHead, sizeof(requestHead),
		"Session: %.8X\r\n"
		"Content-Type: text/parameters\r\n"
		"Content-Length: %d\r\n",
		_remoteSessionId, requestBodyLength);
	assert(requestHeadLength > 0 && requestHeadLength < sizeof(requestHead));

	sendRequestReceiveResponse("SET_PARAMETER", requestHead, requestBody, response);

	return response.statusCode;
}


/**
 * Sends RTSP FLUSH message
 *
 * @return zero on success; non-zero on failure
 */
int RAOPDevice::sendFlush()
{
	int requestHeadLength;
	char requestHead[1024];
	struct Response response;

	requestHeadLength = sprintf_s(requestHead, sizeof(requestHead),
		"Session: %.8X\r\n"
		"RTP-Info: seq=%hu;rtptime=%u\r\n",
		_remoteSessionId, _rtpPacketSequenceNumber, _rtpTime);
	assert(requestHeadLength > 0 && requestHeadLength < sizeof(requestHead));

	sendRequestReceiveResponse("FLUSH", requestHead, "", response);

	return response.statusCode;
}


/**
 * Sends specified RTSP message to remote speakers associated with specified
 * session and retrieves response.  This function requires specified request
 * method to construct the base request headers.  It will also use specified
 * additional headers and specified optional body.  If a body is to be included
 * with the request it is assumed that the caller has included appropriate
 * content type and content length headers.
 *
 * @param requestMethod RTSP request method
 * @param requestHead additional RTSP request headers
 * @param requestBody optional RTSP request body
 * @param response RTSP response structure
 */
void RAOPDevice::sendRequestReceiveResponse(const std::string& requestMethod,
	const std::string& requestHead, const std::string& requestBody,
	Response& response)
{
	const std::string requestText =
		buildRequest(requestMethod, requestHead, requestBody);

	Debugger::print(requestText
		+ "----------------------------------------"
		+ "----------------------------------------");

	sendRequest(requestText);

	// increment sequence number after successful send of message
	++_messageSequenceNumber;

	const std::string responseText = receiveResponse();

	Debugger::print(responseText
		+ "----------------------------------------"
		+ "----------------------------------------");

	parseResponse(responseText, response);

	// validate response sequence number
	assert(response.headers.find("CSeq") != response.headers.end());
	assert(_messageSequenceNumber
		== (unsigned int) atoi(response.headers["CSeq"].c_str()));

	// check response for authenticate header
	if (response.headers.find(WWW_AUTHENTICATE_HEADER) != response.headers.end())
	{
		parseAuthenticateHeader(response.headers[WWW_AUTHENTICATE_HEADER]);
	}
}


std::string RAOPDevice::buildRequest(const std::string& requestMethod,
	const std::string& requestHead, const std::string& requestBody) const
{
	assert(!requestMethod.empty());

	// build request URI
	const std::string requestURI = (_localSessionId == 0 ? "*" : Poco::format(
		"rtsp://%s/%u", _rtspSocket->address().host().toString(), _localSessionId));

	std::string requestText;
	requestText.append(Poco::format("%s %s RTSP/1.0\r\n", requestMethod, requestURI));
	requestText.append(Poco::format("CSeq: %u\r\n", _messageSequenceNumber + 1));
	requestText.append(Poco::format("User-Agent: %s\r\n", Plugin::userAgent()));
	requestText.append(Poco::format("Client-Instance: %016?X\r\n", Plugin::dacpId()));
	requestText.append(Poco::format("DACP-ID: %016?X\r\n", Plugin::dacpId()));
	requestText.append(Poco::format("Active-Remote: %u\r\n", _remoteControlId));
	if (!_authenticationMethod.empty())
		requestText.append(buildAuthorizationHeader(requestMethod, requestURI));
	requestText.append(requestHead);
	requestText.append("\r\n");
	requestText.append(requestBody);

	return requestText;
}


void RAOPDevice::sendRequest(const std::string& requestText)
{
	assert(!requestText.empty());

	const char* const requestBuffer = requestText.c_str();
	const std::size_t requestLength = requestText.length();

	std::size_t bytesSent = 0;
	while (bytesSent < requestLength)
	{
		const int returnCode = _rtspSocket->sendBytes(requestBuffer + bytesSent,
			requestLength - bytesSent);
		if (returnCode <= 0)
		{
			throw std::runtime_error(
				"_rtspSocket->sendBytes returned " + returnCode);
		}

		bytesSent += static_cast<std::size_t>(returnCode);
	}
	if (bytesSent != requestLength)
	{
		throw std::runtime_error("bytesSent != requestLength");
	}
}


std::string RAOPDevice::receiveResponse()
{
	std::vector<char> responseBuffer(8192);
	std::size_t responseLength = 0;

	do
	{
		const int returnCode = _rtspSocket->receiveBytes(
			&responseBuffer[responseLength],
			responseBuffer.size() - responseLength);
		if (returnCode <= 0)
		{
			throw std::runtime_error(
				"_rtspSocket->receiveBytes returned " + returnCode);
		}

		responseLength += static_cast<std::size_t>(returnCode);
	}
	while (std::strncmp(&responseBuffer[responseLength - 4], "\r\n\r\n", 4) != 0);

	assert(responseLength < responseBuffer.size());
	return std::string(&responseBuffer[0], responseLength);
}


void RAOPDevice::parseResponse(const std::string& responseText,
	Response& response)
{
	assert(!responseText.empty());

	std::string::size_type beg = 0;
	std::string::size_type end;

	// parse response for protocol
	end = responseText.find_first_of("/", beg);
	if (end <= beg || end == std::string::npos)
	{
		throw std::invalid_argument("responseText");
	}
	response.protocol = responseText.substr(beg, end - beg);
	beg = responseText.find_first_not_of("/", end);

	// parse response for version
	end = responseText.find_first_of(" ", beg);
	if (end <= beg || end == std::string::npos)
	{
		throw std::invalid_argument("responseText");
	}
	response.version = responseText.substr(beg, end - beg);
	beg = responseText.find_first_not_of(" ", end);

	// parse response for status code
	end = responseText.find_first_of(" ", beg);
	if (end <= beg || end == std::string::npos)
	{
		throw std::invalid_argument("responseText");
	}
	response.statusCode = std::atoi(responseText.substr(beg, end - beg).c_str());
	beg = responseText.find_first_not_of(" ", end);

	// parse response for status text
	end = responseText.find_first_of("\r\n", beg);
	if (end <= beg || end == std::string::npos)
	{
		throw std::invalid_argument("responseText");
	}
	response.statusText = responseText.substr(beg, end - beg);
	beg = responseText.find_first_not_of("\r\n", end);

	// parse response headers
	while (beg != std::string::npos)
	{
		assert(end != std::string::npos);

		// tokenize header into name and value
		end = responseText.find_first_of(":", beg);
		const std::string name = responseText.substr(beg, end - beg);
		beg = responseText.find_first_not_of(": ", end);

		end = responseText.find_first_of("\r\n", beg);
		const std::string value = responseText.substr(beg, end - beg);
		beg = responseText.find_first_not_of("\r\n", end);

		response.headers[name] = value;
	}
}


void RAOPDevice::parseAuthenticateHeader(const std::string& authenticateHeader)
{
	assert(!authenticateHeader.empty());

	// less than zero indicates lower and greater than zero indicates upper;
	// version of the protocol that uses uppercase also happenes to throw a
	// comma into authenticate header to separate realm and nonce, so key off
	// that for now
	_authenticationCasing =
		(authenticateHeader.find(',') != std::string::npos ? 1 : -1);

	std::string::size_type beg = 0;
	std::string::size_type end;

	end = authenticateHeader.find_first_of(" ", beg);
	if (end <= beg || end == std::string::npos)
	{
		throw std::runtime_error("end <= beg || end == std::string::npos");
	}
	_authenticationMethod = authenticateHeader.substr(beg, end - beg);
	beg = authenticateHeader.find_first_not_of(" ", end);

	while (beg != std::string::npos)
	{
		assert(end != std::string::npos);

		end = authenticateHeader.find_first_of("=\"", beg);
		const std::string name = authenticateHeader.substr(beg, end - beg);
		beg = authenticateHeader.find_first_not_of("=\"", end);

		end = authenticateHeader.find_first_of("\"", beg);
		const std::string value = authenticateHeader.substr(beg, end - beg);
		beg = authenticateHeader.find_first_not_of("\", ", end);

		if (name == "realm")
		{
			assert(!value.empty());
			_authenticationRealm = value;
		}
		else if (name == "nonce")
		{
			assert(!value.empty());
			_authenticationNonce = value;
		}
	}
}


#define HEX(bytes, length, string, capacity, casing)                           \
{                                                                              \
	const char *format = (casing < 0 ? "%.2x" : casing > 0 ? "%.2X" : NULL);   \
	for (std::size_t i = 0; i < length; ++i)                                   \
		sprintf_s(string + 2 * i, capacity - 2 * i, format, bytes[i]);         \
}


std::string RAOPDevice::buildAuthorizationHeader(
	const std::string& requestMethod, const std::string& requestURI) const
{
	int length;
	char temp[256];
	byte_t ha1Digest[MD5_DIGEST_LENGTH];
	char ha1String[MD5_DIGEST_LENGTH * 2 + 1];
	byte_t ha2Digest[MD5_DIGEST_LENGTH];
	char ha2String[MD5_DIGEST_LENGTH * 2 + 1];
	byte_t responseDigest[MD5_DIGEST_LENGTH];
	char responseString[MD5_DIGEST_LENGTH * 2 + 1];

	assert(!requestMethod.empty());
	assert(!requestURI.empty());

	// this function assumes RFC 2617 digest authentication
	assert(_authenticationMethod == "Digest");

	// concatenate username, realm and password
	length = sprintf_s(temp, sizeof(temp), "iTunes:%s:%s",
		_authenticationRealm.c_str(), _authenticationPassword.c_str());
	assert(length > 0 && length < sizeof(temp));

	// create HA1 digest
	MD5((unsigned char*) temp, length, ha1Digest);
	HEX(ha1Digest, sizeof(ha1Digest), ha1String, sizeof(ha1String),
		_authenticationCasing);

	// concatenate request method and URI
	length = sprintf_s(temp, sizeof(temp), "%s:%s",
		requestMethod.c_str(), requestURI.c_str());
	assert(length > 0 && length < sizeof(temp));

	// create HA2 digest
	MD5((unsigned char*) temp, length, ha2Digest);
	HEX(ha2Digest, sizeof(ha2Digest), ha2String, sizeof(ha2String),
		_authenticationCasing);

	// concatenate HA1 digest, nonce and HA2 digest
	length = sprintf_s(temp, sizeof(temp), "%s:%s:%s",
		ha1String, _authenticationNonce.c_str(), ha2String);
	assert(length > 0 && length < sizeof(temp));

	// create response digest
	MD5((unsigned char*) temp, length, responseDigest);
	HEX(responseDigest, sizeof(responseDigest), responseString,
		sizeof(responseString), _authenticationCasing);

	// build request authorization header
	length = sprintf_s(temp, sizeof(temp),
		"Authorization: %s username=\"iTunes\", realm=\"%s\", nonce=\"%s\", "
		"uri=\"%s\", response=\"%s\"\r\n",
		_authenticationMethod.c_str(), _authenticationRealm.c_str(),
		_authenticationNonce.c_str(), requestURI.c_str(), responseString);
	assert(length > 0 && length < sizeof(temp));

	return temp;
}
