/* Copyright (c) 2011  Eric Milles <eric.milles@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#ifndef RAOPDevice_h
#define RAOPDevice_h


#include "Device.h"
#include "Platform.h"
#include <map>
#include <memory>
#include <string>
#include <openssl/aes.h>
#include <openssl/rsa.h>
#include <Poco/Net/StreamSocket.h>


enum AudioJackStatus
{
	AUDIO_JACK_DISCONNECTED = 0,
	AUDIO_JACK_CONNECTED = 1
};


class RAOPDevice
:
	public Device
{
public:
	static void staticFree();
	static void staticInit();
	static void staticReset();
	static const OutputFormat& audioFormat();
	static std::size_t audioSendLimit();
	static void encodeAudio(const void*, std::size_t, const void*&, std::size_t&);
	static void sentAudio(std::size_t);

private:
	/** RAOP RSA encryption public key */
	static RSA* _rsaKey;

	/** RAOP AES encryption key */
	static AES_KEY _aesKey;
	static char _encodedKey[384];

	/** RAOP AES encryption initialization vector */
	static byte_t _aesIV[16];
	static char _encodedVector[32];

	/** RTP packet sequence number */
	static unsigned short _rtpPacketSequenceNumber;

	/** RTP time */
	static unsigned int _rtpTime;

	/** RTP synchronization source identifier */
	static unsigned int _rtpSynchronizationSourceId;

public:
	RAOPDevice();
	~RAOPDevice();

	bool isOpen() const;
	void open();
	void close();

	void setBalance(double);
	void setVolume(double);

	unsigned int remoteControlId() const;

	std::time_t latency(const OutputFormat&) const;
	std::size_t buffered() const;
	std::size_t canWrite() const;

	void write(const void*, std::size_t);
	void flush();
	void reset();

///
	int initiate(Poco::Net::StreamSocket&, bool);
	int negotiate(Poco::Net::StreamSocket&, AudioJackStatus&);
	void setPassword(const std::string&);
///

private:
	void reinit();

	int sendOptions(bool);
	int sendAnnounce();
	int sendSetup(AudioJackStatus&);
	int sendRecord();
	int sendTeardown();
	int sendSetParameter(const std::string&, const std::string&);
	int sendFlush();

	struct Response
	{
		std::string protocol;
		std::string version;
		int statusCode;
		std::string statusText;
		std::map<std::string,std::string> headers;
	};

	void sendRequestReceiveResponse(const std::string&, const std::string&, const std::string&, Response&);
	std::string buildRequest(const std::string&, const std::string&, const std::string&) const;
	void sendRequest(const std::string&);
	std::string receiveResponse();
	void parseResponse(const std::string&, Response&);
	void parseAuthenticateHeader(const std::string&);
	std::string buildAuthorizationHeader(const std::string&, const std::string&) const;

	/** socket used to connect to remote speakers for RTSP */
	std::auto_ptr<Poco::Net::StreamSocket> _rtspSocket;

	/** socket used to connect to remote speakers for RTP */
	std::auto_ptr<Poco::Net::StreamSocket> _rtpSocket;

	/** is RTSP teardown message needed on disconnect */
	bool _teardownRequired;

	/** audio latency of remote speakers */
	unsigned int _audioLatency;

	/** RTSP message sequence number */
	unsigned int _messageSequenceNumber;

	/** RTSP local session identifier */
	unsigned int _localSessionId;

	/** RTSP remote session identifier */
	unsigned int _remoteSessionId;

	/** DACP remote control identifier */
	unsigned int _remoteControlId;

	/** port used for RTP audio traffic */
	unsigned short _serverPort;

	/** RTSP authentication casing */
	int _authenticationCasing;

	/** RTSP authentication method */
	std::string _authenticationMethod;

	/** RTSP authentication realm */
	std::string _authenticationRealm;

	/** RTSP authentication nonce */
	std::string _authenticationNonce;

	/** RTSP authentication password */
	std::string _authenticationPassword;
};


#endif // RAOPDevice_h
