/* Copyright (c) 2011  Eric Milles <eric.milles@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "Debugger.h"
#include "Platform.h"
#include "Plugin.h"
#include "RemoteControl.h"
#include <algorithm>
#include <cassert>
#include <cstdio>
#include <cstring>
#include <map>
#include <stdexcept>
#include <string>
#include <vector>
#include <Poco/DateTimeFormat.h>
#include <Poco/DateTimeFormatter.h>
#include <Poco/Format.h>
#include <Poco/StringTokenizer.h>
#include <Poco/Net/ServerSocket.h>
#include <Poco/Net/StreamSocket.h>


using Poco::DateTime;
using Poco::DateTimeFormat;
using Poco::DateTimeFormatter;
using Poco::StringTokenizer;
using Poco::Thread;
using Poco::Timespan;
using Poco::Net::ServerSocket;
using Poco::Net::Socket;
using Poco::Net::StreamSocket;


enum Command
{
	UNKNOWN,
	PLAY,
	PAUSE,
	STOP,
	SKIP_FWD,
	SKIP_REV,
	SEEK_FWD,
	SEEK_REV,
	VOLUME_UP,
	VOLUME_DOWN,
	MUTE,
	SHUFFLE
};


struct Request
{
	Command command;
	std::map<std::string,std::string> headers;
};


static const unsigned short DACP_PORT = 3689;

static const std::string REMOTE_CONTROL_ID_PARAMETER("Active-Remote");


//------------------------------------------------------------------------------


static std::string receiveRequest(StreamSocket& clientSocket)
{
	std::vector<char> requestBuffer(4096);
	int requestLength = 0;

	do
	{
		const int returnCode = clientSocket.receiveBytes(
			&requestBuffer[requestLength], requestBuffer.size() - requestLength);
		if (returnCode <= 0)
		{
			throw std::runtime_error(
				"clientSocket.receiveBytes returned " + returnCode);
		}

		requestLength += returnCode;
	}
	while (std::strncmp(&requestBuffer[requestLength - 4], "\r\n\r\n", 4) != 0);

	assert(requestLength < requestBuffer.size());
	return std::string(&requestBuffer[0], requestLength);
}


static void parseRequest(const std::string& requestText, Request& request)
{
	const StringTokenizer requestLines(requestText, "\r\n",
		StringTokenizer::TOK_IGNORE_EMPTY);
	StringTokenizer::Iterator requestLinesIterator = requestLines.begin();
	if (requestLinesIterator == requestLines.end())
	{
		throw std::invalid_argument("requestText");
	}

	// parse request for method
	const StringTokenizer requestTokens(*requestLinesIterator, " ");
	StringTokenizer::Iterator requestTokensIterator = requestTokens.begin();
	if (requestTokensIterator == requestTokens.end()
		|| *requestTokensIterator != "GET")
	{
		throw std::invalid_argument("requestText");
	}

	// parse request for URI
	++requestTokensIterator;
	if (requestLinesIterator == requestLines.end())
	{
		throw std::invalid_argument("requestText");
	}

	// parse URI for command
	const StringTokenizer uriTokens(*requestTokensIterator, "/",
		StringTokenizer::TOK_IGNORE_EMPTY);
	request.command = UNKNOWN;
	if (uriTokens[0] == "ctrl-int")
	{
		if (uriTokens.count() != 3)
		{
			throw std::invalid_argument("requestText");
		}

		if (uriTokens[1] == "1")
		{
			if (uriTokens[2] == "playpause")
			{
				request.command = PLAY;
			}
			else if (uriTokens[2] == "pause")
			{
				request.command = PAUSE;
			}
			else if (uriTokens[2] == "stop")
			{
				request.command = STOP;
			}
			else if (uriTokens[2] == "nextitem")
			{
				request.command = SKIP_FWD;
			}
			else if (uriTokens[2] == "previtem")
			{
				request.command = SKIP_REV;
			}
			else if (uriTokens[2] == "beginff")
			{
				request.command = SEEK_FWD;
			}
			else if (uriTokens[2] == "beginrew")
			{
				request.command = SEEK_REV;
			}
			else if (uriTokens[2] == "volumeup")
			{
				request.command = VOLUME_UP;
			}
			else if (uriTokens[2] == "volumedown")
			{
				request.command = VOLUME_DOWN;
			}
			else if (uriTokens[2] == "mutetoggle")
			{
				request.command = MUTE;
			}
			else if (uriTokens[2] == "shufflesongs")
			{
				request.command = SHUFFLE;
			}
		}
	}

	// parse request for protocol
	++requestTokensIterator;
	if (requestTokensIterator == requestTokens.end()
		|| *requestTokensIterator != "HTTP/1.1")
	{
		throw std::invalid_argument("requestText");
	}

	// parse request for headers
	while (++requestLinesIterator != requestLines.end())
	{
		// tokenize header into name and value
		const StringTokenizer headerTokens(*requestLinesIterator, ":");

		// skip space at beginning of value string
		request.headers[headerTokens[0]] = headerTokens[1].substr(1);
	}
}


static std::string buildResponse(const bool requestUnderstood)
{
	std::string responseText;
	responseText.append((requestUnderstood
		? "HTTP/1.1 204 No Content\r\n" : "HTTP/1.1 501 Not Implemented\r\n"));
	responseText.append(Poco::format("Date: %s\r\n",
		DateTimeFormatter::format(DateTime(), DateTimeFormat::HTTP_FORMAT)));
	responseText.append(Poco::format("DAAP-Server: %s\r\n",
		Plugin::userAgent()));
	responseText.append("Content-Type: application/x-dmap-tagged\r\n");
	responseText.append("Content-Length: 0\r\n");
	responseText.append("\r\n");

	return responseText;
}


static void sendResponse(const std::string& responseText, StreamSocket& socket)
{
	assert(!responseText.empty());

	const char* const response = responseText.c_str();
	const std::size_t responseLength = responseText.length();

	// send response on specified socket
	std::size_t bytesSent = 0;
	while (bytesSent < responseLength)
	{
		int returnCode = socket.sendBytes(
			response + bytesSent, responseLength - bytesSent);
		if (returnCode <= 0)
		{
			throw std::runtime_error("socket.sendBytes returned " + returnCode);
		}

		bytesSent += static_cast<std::size_t>(returnCode);
	}
	if (bytesSent != responseLength)
	{
		throw std::runtime_error("bytesSent != responseLength");
	}
}


static void controlPlayer(Player& player, Command command)
{
	switch (command)
	{
	case PLAY:
		player.play();
		break;
	case PAUSE:
		player.pause();
		break;
	case STOP:
		player.stop();
		break;
	case SKIP_FWD:
		player.skipForward();
		break;
	case SKIP_REV:
		player.skipBackward();
		break;
	case SEEK_FWD:
		player.seekForward();
		break;
	case SEEK_REV:
		player.seekBackward();
		break;
	case VOLUME_UP:
		player.increaseVolume();
		break;
	case VOLUME_DOWN:
		player.decreaseVolume();
		break;
	case MUTE:
		player.mute();
		break;
	case SHUFFLE:
		player.shuffle();
		break;
	default:
		break;
	}
}


static void handleRequest(StreamSocket clientSocket,
	const DeviceManager& deviceManager, Player& player)
{
	const std::string requestText = receiveRequest(clientSocket);

	Debugger::print(requestText
		+ "----------------------------------------"
		+ "----------------------------------------");

	Request request;
	parseRequest(requestText, request);

	// validate remote control identifier before allowing requested remote
	// control command to be executed
	unsigned int id;
	if (request.headers.find(REMOTE_CONTROL_ID_PARAMETER) == request.headers.end()
		|| !std::sscanf(request.headers[REMOTE_CONTROL_ID_PARAMETER].c_str(), "%u", &id)
		|| !deviceManager.isValidRemoteControlId(id))
	{
		request.command = UNKNOWN;
	}

	const std::string responseText = buildResponse(request.command != UNKNOWN);

	Debugger::print(responseText
		+ "----------------------------------------"
		+ "----------------------------------------");

	sendResponse(responseText, clientSocket);

	controlPlayer(player, request.command);
}


//------------------------------------------------------------------------------


RemoteControl::RemoteControl(DeviceManager& deviceManager, Player& player)
:
	_deviceManager(deviceManager),
	_player(player),
	_registerOperation(0),
	_stopThread(false)
{
	_thread.start(*this);
}


RemoteControl::~RemoteControl()
{
	try
	{
		_stopThread = true;
		_thread.join();

		if (_registerOperation)
		{
			ServiceDiscovery::stop(_registerOperation);
		}
	}
	catch (...)
	{
	}
}


void RemoteControl::run()
{
	Debugger::setThreadName("RemoteControl::run");

	try
	{
		// create socket to listen for DACP requests
		ServerSocket serverSocket;
		serverSocket.bind(DACP_PORT);
		serverSocket.setNoDelay(true);
		serverSocket.setOption(SOL_SOCKET, SO_EXCLUSIVEADDRUSE, TRUE);
		serverSocket.listen();

		// advertise DACP server
		registerService();

		Socket::SocketList clientList;

		Socket::SocketList readList;
		Socket::SocketList writeList;
		Socket::SocketList exceptList;
		const Timespan timeout(0, 20000);

		while (!_stopThread)
		{
			Thread::yield();

			readList.clear();
			readList.push_back(serverSocket);
			readList.insert(readList.end(), clientList.begin(), clientList.end());

			// select sockets in set which can be read from without blocking
			if (Socket::select(readList, writeList, exceptList, timeout) > 0)
			{
				// check if server socket is selected
				const Socket::SocketList::iterator pos =
					std::find(readList.begin(), readList.end(), serverSocket);
				if (pos != readList.end())
				{
					readList.erase(pos);

					// accept new client connection
					clientList.push_back(serverSocket.acceptConnection());
				}

				// check if any client sockets are selected
				if (!readList.empty())
				{
					for (Socket::SocketList::iterator it = readList.begin();
						it != readList.end(); ++it)
					{
						try
						{
							handleRequest(*it, _deviceManager, _player);
						}
						catch (...)
						{
							(*it).close();
							clientList.erase(
								std::find(clientList.begin(), clientList.end(), *it));
						}
					}
				}
			}
		}

		// disconnect any active client connections
		for (Socket::SocketList::iterator it = clientList.begin(),
			itEnd = clientList.end(); it != itEnd; ++it)
		{
			(*it).close();
		}
	}
	catch (...)
	{
	}
}


void RemoteControl::registerService()
{
	if (ServiceDiscovery::isAvailable())
	{
		const std::string dacpId(Poco::format("%016?X", Plugin::dacpId()));
		const std::string serviceName("iTunes_Ctrl_" + dacpId);

		ServiceDiscovery::TXTRecord txtRecord;
		txtRecord.put("Ver", "65536");
		txtRecord.put("DbId", dacpId);

		_registerOperation = ServiceDiscovery::registerService(
			serviceName, "_dacp._tcp.", DACP_PORT, txtRecord, *this);
		ServiceDiscovery::start(_registerOperation);
	}
}


void RemoteControl::serviceRegistered(DNSServiceRef serviceRef,
	const std::string name, const std::string type)
{
	assert(serviceRef == _registerOperation);
}
