/* Copyright (c) 2011  Eric Milles <eric.milles@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "Debugger.h"
#include "ServiceDiscovery.h"
#include "Uncopyable.h"
#include <algorithm>
#include <cassert>
#include <limits>
#include <stdexcept>
#include <vector>
#include <winsock2.h>
#include <Poco/ByteOrder.h>
#include <Poco/Mutex.h>
#include <Poco/Runnable.h>
#include <Poco/ScopedLock.h>
#include <Poco/ScopedUnlock.h>
#include <Poco/SharedLibrary.h>
#include <Poco/Thread.h>


using Poco::ByteOrder;
using Poco::FastMutex;
using Poco::Runnable;
using Poco::ScopedLock;
using Poco::ScopedUnlock;
using Poco::SharedLibrary;
using Poco::Thread;


class ServiceDiscoveryImpl
:
	public Runnable,
	private Uncopyable
{
	friend class ServiceDiscovery;

	ServiceDiscoveryImpl();
	~ServiceDiscoveryImpl();

	void start(DNSServiceRef);
	void stop(DNSServiceRef);
	void run();

	static void DNSSD_API browseCallback(
		DNSServiceRef,
		DNSServiceFlags,
		uint32_t,
		DNSServiceErrorType,
		const char*,
		const char*,
		const char*,
		void*);
	static void DNSSD_API queryCallback(
		DNSServiceRef,
		DNSServiceFlags,
		uint32_t,
		DNSServiceErrorType,
		const char*,
		uint16_t,
		uint16_t,
		uint16_t,
		const void*,
		uint32_t,
		void*);
	static void DNSSD_API registerCallback(
		DNSServiceRef,
		DNSServiceFlags,
		DNSServiceErrorType,
		const char*,
		const char*,
		const char*,
		void*);
	static void DNSSD_API resolveCallback(
		DNSServiceRef,
		DNSServiceFlags,
		uint32_t,
		DNSServiceErrorType,
		const char*,
		const char*,
		uint16_t,
		uint16_t,
		const unsigned char*,
		void*);

	typedef DNSServiceErrorType (DNSSD_API *DNSServiceGetPropertyProc)(
		char*,
		void*,
		uint32_t*);
	typedef DNSServiceErrorType (DNSSD_API *DNSServiceBrowseProc)(
		DNSServiceRef*,
		DNSServiceFlags,
		uint32_t,
		const char*,
		const char*,
		DNSServiceBrowseReply,
		void*);
	typedef DNSServiceErrorType (DNSSD_API *DNSServiceQueryRecordProc)(
		DNSServiceRef*,
		DNSServiceFlags,
		uint32_t,
		const char*,
		uint16_t,
		uint16_t,
		DNSServiceQueryRecordReply,
		void*);
	typedef DNSServiceErrorType (DNSSD_API *DNSServiceRegisterProc)(
		DNSServiceRef*,
		DNSServiceFlags,
		uint32_t,
		const char*,
		const char*,
		const char*,
		const char*,
		uint16_t,
		uint16_t,
		const void*,
		DNSServiceRegisterReply,
		void*);
	typedef DNSServiceErrorType (DNSSD_API *DNSServiceResolveProc)(
		DNSServiceRef*,
		DNSServiceFlags,
		uint32_t,
		const char*,
		const char*,
		const char*,
		DNSServiceResolveReply,
		void*);
	typedef int (DNSSD_API *DNSServiceConstructFullNameProc)(
		char*,
		const char*,
		const char*,
		const char*);
	typedef void (DNSSD_API *DNSServiceRefDeallocateProc)(
		DNSServiceRef);
	typedef int (DNSSD_API *DNSServiceRefSockFDProc)(
		DNSServiceRef);
	typedef DNSServiceErrorType (DNSSD_API *DNSServiceProcessResultProc)(
		DNSServiceRef);
	typedef void (DNSSD_API *TXTRecordCreateProc)(
		TXTRecordRef*,
		uint16_t,
		void*);
	typedef void (DNSSD_API *TXTRecordDeallocateProc)(
		TXTRecordRef*);
	typedef uint16_t (DNSSD_API *TXTRecordGetLengthProc)(
		const TXTRecordRef*);
	typedef const void* (DNSSD_API *TXTRecordGetBytesPtrProc)(
		const TXTRecordRef*);
	typedef int (DNSSD_API *TXTRecordContainsKeyProc)(
		uint16_t,
		const void*,
		const char*);
	typedef uint16_t (DNSSD_API *TXTRecordCountKeysProc)(
		uint16_t,
		const void*);
	typedef DNSServiceErrorType (DNSSD_API *TXTRecordGetItemAtIndexProc)(
		uint16_t,
		const void*,
		uint16_t,
		uint16_t,
		char*,
		uint8_t*,
		const void**);
	typedef const void* (DNSSD_API *TXTRecordGetValuePtrProc)(
		uint16_t,
		const void*,
		const char*,
		uint8_t*);
	typedef DNSServiceErrorType (DNSSD_API *TXTRecordSetValueProc)(
		TXTRecordRef*,
		const char*,
		uint8_t,
		const void*);

	SharedLibrary _sharedLibrary;

	DNSServiceGetPropertyProc _getProperty;
	DNSServiceBrowseProc _browseServices;
	DNSServiceQueryRecordProc _queryService;
	DNSServiceRegisterProc _registerService;
	DNSServiceResolveProc _resolveService;
	DNSServiceConstructFullNameProc _makeFullName;
	DNSServiceRefDeallocateProc _serviceRefDeallocate;
	DNSServiceRefSockFDProc _serviceRefGetSocketFD;
	DNSServiceProcessResultProc _serviceRefProcessResult;
	TXTRecordCreateProc _txtRecordCreate;
	TXTRecordDeallocateProc _txtRecordDeallocate;
	TXTRecordGetLengthProc _txtRecordGetLength;
	TXTRecordGetBytesPtrProc _txtRecordGetBytesPtr;
	TXTRecordContainsKeyProc _txtRecordContainsKey;
	TXTRecordCountKeysProc _txtRecordCountKeys;
	TXTRecordGetItemAtIndexProc _txtRecordGetItemAtIndex;
	TXTRecordGetValuePtrProc _txtRecordGetValue;
	TXTRecordSetValueProc _txtRecordSetValue;

	std::vector<DNSServiceRef> _activeServiceRefs;

	typedef FastMutex Mutex;
	Mutex _mutex;

	Thread _thread;
};


//------------------------------------------------------------------------------


ServiceDiscoveryImpl& ServiceDiscovery::impl()
{
	static ServiceDiscoveryImpl singleton;
	return singleton;
}


bool ServiceDiscovery::isAvailable()
{
	try
	{
		// check if SD daemon is running by requesting its version property
		uint32_t versionNo = 0;
		uint32_t versionSz = 4;
		const DNSServiceErrorType error = impl()._getProperty(
			kDNSServiceProperty_DaemonVersion,
			&versionNo,
			&versionSz);
		if (error != kDNSServiceErr_NoError)
		{
			throw std::runtime_error("_getProperty returned error code " + error);
		}
		assert(versionNo > 0 && versionSz == 4);

		return true;
	}
	catch (...)
	{
		return false;
	}
}


void ServiceDiscovery::start(DNSServiceRef serviceRef)
{
	try
	{
		impl().start(serviceRef);
	}
	catch (...)
	{
		impl()._serviceRefDeallocate(serviceRef);

		throw;
	}
}


void ServiceDiscovery::stop(DNSServiceRef serviceRef)
{
	try
	{
		impl().stop(serviceRef);
	}
	catch (...)
	{
		impl()._serviceRefDeallocate(serviceRef);

		throw;
	}
}


std::string ServiceDiscovery::fullName(const std::string& name,
	const std::string& type)
{
	char fullName[kDNSServiceMaxDomainName];
	const int error = impl()._makeFullName(
		fullName,
		name.c_str(),
		type.c_str(),
		"local.");
	if (error != 0)
	{
		throw std::runtime_error("_makeFullName returned error code " + error);
	}
	return fullName;
}


DNSServiceRef ServiceDiscovery::browseServices(const std::string& type,
	BrowseListener& listener)
{
	DNSServiceRef serviceRef;
	const DNSServiceErrorType error = impl()._browseServices(
		&serviceRef,
		0, // flags
		0, // interface index
		type.c_str(),
		"local.",
		ServiceDiscoveryImpl::browseCallback,
		(void*) &listener);
	if (error != kDNSServiceErr_NoError)
	{
		throw std::runtime_error("_browseServices returned error code " + error);
	}
	return serviceRef;
}


DNSServiceRef ServiceDiscovery::queryService(const std::string& fullName,
	const uint16_t recordType, QueryListener& listener)
{
	DNSServiceRef serviceRef;
	const DNSServiceErrorType error = impl()._queryService(
		&serviceRef,
		0, // flags
		0, // interface index
		fullName.c_str(),
		recordType,
		kDNSServiceClass_IN,
		ServiceDiscoveryImpl::queryCallback,
		(void*) &listener);
	if (error != kDNSServiceErr_NoError)
	{
		throw std::runtime_error("_queryService returned error code " + error);
	}
	return serviceRef;
}


DNSServiceRef ServiceDiscovery::registerService(const std::string& name,
	const std::string& type, const uint16_t port, const TXTRecord& txtRecord,
	RegisterListener& listener)
{
	DNSServiceRef serviceRef;
	const DNSServiceErrorType error = impl()._registerService(
		&serviceRef,
		0, // flags
		0, // interface index
		name.c_str(),
		type.c_str(),
		"local.",
		NULL, // host
		ByteOrder::toNetwork(port),
		txtRecord.len(),
		txtRecord.buf(),
		ServiceDiscoveryImpl::registerCallback,
		(void*) &listener);
	if (error != kDNSServiceErr_NoError)
	{
		throw std::runtime_error("_registerService returned error code " + error);
	}
	return serviceRef;
}


DNSServiceRef ServiceDiscovery::resolveService(const std::string& name,
	const std::string& type, ResolveListener& listener)
{
	DNSServiceRef serviceRef;
	const DNSServiceErrorType error = impl()._resolveService(
		&serviceRef,
		0, //flags
		0, // interface index
		name.c_str(),
		type.c_str(),
		"local.",
		ServiceDiscoveryImpl::resolveCallback,
		(void*) &listener);
	if (error != kDNSServiceErr_NoError)
	{
		throw std::runtime_error("_resolveService returned error code " + error);
	}
	return serviceRef;
}


//------------------------------------------------------------------------------


void DNSSD_API ServiceDiscoveryImpl::browseCallback(
	DNSServiceRef serviceRef,
	const DNSServiceFlags flags,
	const uint32_t /*interfaceIndex*/,
	const DNSServiceErrorType error,
	const char* const name,
	const char* const type,
	const char* const domain,
	void* const context)
{
	try
	{
		assert(error == kDNSServiceErr_NoError);
		assert(std::string(domain) == "local.");

		if (flags & kDNSServiceFlagsAdd)
		{
			static_cast<ServiceDiscovery::BrowseListener*>(context)->
				serviceFound(serviceRef, name, type);
		}
		else
		{
			static_cast<ServiceDiscovery::BrowseListener*>(context)->
				serviceLost(serviceRef, name, type);
		}
	}
	catch (...)
	{
	}
}


void DNSSD_API ServiceDiscoveryImpl::queryCallback(
	DNSServiceRef serviceRef,
	const DNSServiceFlags flags,
	const uint32_t /*interfaceIndex*/,
	const DNSServiceErrorType error,
	const char* const fullName,
	const uint16_t rrtype,
	const uint16_t rrclass,
	const uint16_t rdlen,
	const void* const rdata,
	const uint32_t ttl,
	void* const context)
{
	try
	{
		assert(flags & kDNSServiceFlagsAdd);
		assert(error == kDNSServiceErr_NoError);
		assert(rrclass == kDNSServiceClass_IN);

		static_cast<ServiceDiscovery::QueryListener*>(context)->
			serviceQueried(serviceRef, fullName, rrtype, rdlen, rdata, ttl);
	}
	catch (...)
	{
	}
}


void DNSSD_API ServiceDiscoveryImpl::registerCallback(
	DNSServiceRef serviceRef,
	const DNSServiceFlags /*flags*/,
	const DNSServiceErrorType error,
	const char* const name,
	const char* const type,
	const char* const domain,
	void* const context)
{
	try
	{
		assert(error == kDNSServiceErr_NoError);
		assert(std::string(domain) == "local.");

		static_cast<ServiceDiscovery::RegisterListener*>(context)->
			serviceRegistered(serviceRef, name, type);
	}
	catch (...)
	{
	}
}


void DNSSD_API ServiceDiscoveryImpl::resolveCallback(
	DNSServiceRef serviceRef,
	const DNSServiceFlags /*flags*/,
	const uint32_t /*interfaceIndex*/,
	const DNSServiceErrorType error,
	const char* const fullName,
	const char* const host,
	const uint16_t port,
	const uint16_t txtLength,
	const unsigned char* const txtRecord,
	void* const context)
{
	try
	{
		assert(error == kDNSServiceErr_NoError);

		static_cast<ServiceDiscovery::ResolveListener*>(context)->
			serviceResolved(serviceRef, fullName, host, ByteOrder::fromNetwork(port),
				ServiceDiscovery::TXTRecord(txtRecord, txtLength));
	}
	catch (...)
	{
	}
}


//------------------------------------------------------------------------------


ServiceDiscoveryImpl::ServiceDiscoveryImpl()
:
	_sharedLibrary("dnssd.dll"),
	_getProperty(static_cast<DNSServiceGetPropertyProc>(
		_sharedLibrary.getSymbol("DNSServiceGetProperty"))),
	_browseServices(static_cast<DNSServiceBrowseProc>(
		_sharedLibrary.getSymbol("DNSServiceBrowse"))),
	_queryService(static_cast<DNSServiceQueryRecordProc>(
		_sharedLibrary.getSymbol("DNSServiceQueryRecord"))),
	_registerService(static_cast<DNSServiceRegisterProc>(
		_sharedLibrary.getSymbol("DNSServiceRegister"))),
	_resolveService(static_cast<DNSServiceResolveProc>(
		_sharedLibrary.getSymbol("DNSServiceResolve"))),
	_makeFullName(static_cast<DNSServiceConstructFullNameProc>(
		_sharedLibrary.getSymbol("DNSServiceConstructFullName"))),
	_serviceRefDeallocate(static_cast<DNSServiceRefDeallocateProc>(
		_sharedLibrary.getSymbol("DNSServiceRefDeallocate"))),
	_serviceRefGetSocketFD(static_cast<DNSServiceRefSockFDProc>(
		_sharedLibrary.getSymbol("DNSServiceRefSockFD"))),
	_serviceRefProcessResult(static_cast<DNSServiceProcessResultProc>(
		_sharedLibrary.getSymbol("DNSServiceProcessResult"))),
	_txtRecordCreate(static_cast<TXTRecordCreateProc>(
		_sharedLibrary.getSymbol("TXTRecordCreate"))),
	_txtRecordDeallocate(static_cast<TXTRecordDeallocateProc>(
		_sharedLibrary.getSymbol("TXTRecordDeallocate"))),
	_txtRecordGetLength(static_cast<TXTRecordGetLengthProc>(
		_sharedLibrary.getSymbol("TXTRecordGetLength"))),
	_txtRecordGetBytesPtr(static_cast<TXTRecordGetBytesPtrProc>(
		_sharedLibrary.getSymbol("TXTRecordGetBytesPtr"))),
	_txtRecordContainsKey(static_cast<TXTRecordContainsKeyProc>(
		_sharedLibrary.getSymbol("TXTRecordContainsKey"))),
	_txtRecordCountKeys(static_cast<TXTRecordCountKeysProc>(
		_sharedLibrary.getSymbol("TXTRecordGetCount"))),
	_txtRecordGetItemAtIndex(static_cast<TXTRecordGetItemAtIndexProc>(
		_sharedLibrary.getSymbol("TXTRecordGetItemAtIndex"))),
	_txtRecordGetValue(static_cast<TXTRecordGetValuePtrProc>(
		_sharedLibrary.getSymbol("TXTRecordGetValuePtr"))),
	_txtRecordSetValue(static_cast<TXTRecordSetValueProc>(
		_sharedLibrary.getSymbol("TXTRecordSetValue")))
{
}


ServiceDiscoveryImpl::~ServiceDiscoveryImpl()
{
	try
	{
		while (!_activeServiceRefs.empty())
		{
			stop(_activeServiceRefs.front());
		}

		_thread.join(500);
	}
	catch (...)
	{
	}
}


void ServiceDiscoveryImpl::start(DNSServiceRef serviceRef)
{
	assert(serviceRef != 0);

	ScopedLock<Mutex> lock(_mutex);

	_activeServiceRefs.push_back(serviceRef);

	if (!_thread.isRunning())
	{
		_thread.start(*this);
	}
}


void ServiceDiscoveryImpl::stop(DNSServiceRef serviceRef)
{
	assert(serviceRef != 0);

	ScopedLock<Mutex> lock(_mutex);

	std::vector<DNSServiceRef>::iterator pos = std::find(
		_activeServiceRefs.begin(), _activeServiceRefs.end(), serviceRef);
	assert(pos != _activeServiceRefs.end());
	_activeServiceRefs.erase(pos);

	_serviceRefDeallocate(serviceRef);
}


void ServiceDiscoveryImpl::run()
{
	Debugger::setThreadName("ServiceDiscoveryImpl::run");

	const timeval timeout = { 0, 10000 };

	fd_set readfds;

	for (;;)
	{
		Thread::yield();

		ScopedLock<Mutex> lock(_mutex);

		// check stopping condition after acquiring mutex
		if (_activeServiceRefs.empty())
			break;

		FD_ZERO(&readfds);
		assert(_activeServiceRefs.size() <= FD_SETSIZE);
		for (std::vector<DNSServiceRef>::const_iterator it =
			_activeServiceRefs.begin(); it != _activeServiceRefs.end(); ++it)
		{
			assert(*it != 0);
#pragma warning(push)
#pragma warning(disable:4127)
#pragma warning(disable:4389)
			FD_SET(_serviceRefGetSocketFD(*it), &readfds);
#pragma warning(pop)
		}

		// find active service discovery ops with available results
		const int returnCode = select(0, &readfds, NULL, NULL, &timeout);
		if (returnCode == SOCKET_ERROR)
		{
			const int lastError = WSAGetLastError();
			assert(lastError == 0);
		}
		else if (returnCode > 0)
		{
			// check each active service discovery op to see if it was selected
			for (int r = 0; r < _activeServiceRefs.size(); ++r)
			{
				assert(_activeServiceRefs[r] != 0);
				if (FD_ISSET(
					_serviceRefGetSocketFD(_activeServiceRefs[r]), &readfds))
				{
					// unlock mutex because callback client may wish to start or
					// stop one or more service discovery operations; a deadlock
					// would occur if the mutex remained locked
					ScopedUnlock<Mutex> unlock(_mutex);

					// process next service discovery result
					const DNSServiceErrorType error =
						_serviceRefProcessResult(_activeServiceRefs[r]);
					if (error != kDNSServiceErr_NoError
						&& error != kDNSServiceErr_Unknown)
					{
						assert(error == kDNSServiceErr_NoError
							|| error == kDNSServiceErr_Unknown);
					}
				}
			}
		}
	}
}


//------------------------------------------------------------------------------


ServiceDiscovery::TXTRecord::TXTRecord(const void* const buf, const uint16_t len)
:
	_buf(buf),
	_len(len)
{
}


ServiceDiscovery::TXTRecord::TXTRecord()
:
	_buf(0),
	_len(0)
{
	ServiceDiscovery::impl()._txtRecordCreate(&_ref, 0, NULL);
	put("txtvers", "1");
}


ServiceDiscovery::TXTRecord::~TXTRecord()
{
	if (!_buf)
		ServiceDiscovery::impl()._txtRecordDeallocate(&_ref);
}


const void* ServiceDiscovery::TXTRecord::buf() const
{
	if (_buf)
		return _buf;

	return ServiceDiscovery::impl()._txtRecordGetBytesPtr(&_ref);
}


uint16_t ServiceDiscovery::TXTRecord::len() const
{
	if (_len)
		return _len;

	return ServiceDiscovery::impl()._txtRecordGetLength(&_ref);
}


std::string ServiceDiscovery::TXTRecord::str() const
{
	const void* const buf = this->buf();
	const uint16_t len = this->len();
	const uint16_t num = ServiceDiscovery::impl()._txtRecordCountKeys(len, buf);
	std::vector<char> key(std::numeric_limits<uint8_t>::max() + 1);
	std::string str;

	for (uint16_t idx = 0; idx < num; ++idx)
	{
		uint8_t siz;
		const void* val;
		const DNSServiceErrorType error =
			ServiceDiscovery::impl()._txtRecordGetItemAtIndex(len, buf, idx,
				static_cast<uint16_t>(key.size()), &key[0], &siz, &val);
		if (error != kDNSServiceErr_NoError)
		{
			throw std::runtime_error(
				"_txtRecordGetItemAtIndex returned error code " + error);
		}

		if (idx > 0)
			str.append(" ");
		str.append(&key[0]);
		str.append("=");
		str.append(static_cast<const char*>(val), siz);
	}
	return str;
}


bool ServiceDiscovery::TXTRecord::has(const std::string& key) const
{
	return (0 != ServiceDiscovery::impl()._txtRecordContainsKey(
		len(), buf(), key.c_str()));
}


const void* ServiceDiscovery::TXTRecord::get(const std::string& key,
	uint8_t& len) const
{
	return ServiceDiscovery::impl()._txtRecordGetValue(
		this->len(), buf(), key.c_str(), &len);
}


std::string ServiceDiscovery::TXTRecord::get(const std::string& key) const
{
	uint8_t len;
	const void* val = get(key, len);
	return std::string(static_cast<const char*>(val), len);
}


void ServiceDiscovery::TXTRecord::put(const std::string& key, const void* val,
	const uint8_t len)
{
	const DNSServiceErrorType error =
		ServiceDiscovery::impl()._txtRecordSetValue(&_ref, key.c_str(), len, val);
	if (error != kDNSServiceErr_NoError)
	{
		throw std::runtime_error("_txtRecordSetValue returned error code " + error);
	}
}


void ServiceDiscovery::TXTRecord::put(const std::string& key, const std::string& val)
{
	if (val.length() > std::numeric_limits<uint8_t>::max())
	{
		throw std::invalid_argument(
			"val.length() > std::numeric_limits<uint8_t>::max()");
	}
	put(key, val.c_str(), static_cast<uint8_t>(val.length()));
}
