/* Copyright (c) 2011  Eric Milles <eric.milles@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "ConnectDialog.h"
#include "Dialog.h"
#include "Platform.h"
#include "Resources.h"
#include "ServiceDiscovery.h"
#include <cassert>
#include <cmath>
#include <commctrl.h>
#include <Poco/Net/SocketAddress.h>


using Poco::Timespan;
using Poco::Net::IPAddress;
using Poco::Net::SocketAddress;
using Poco::Net::StreamSocket;


class ConnectDialogImpl
:
	public Dialog,
	public ServiceDiscovery::ResolveListener,
	public ServiceDiscovery::QueryListener,
	private Uncopyable
{
	friend class ConnectDialog;

	explicit ConnectDialogImpl(const DeviceInfo&);

	void onInitialize();
	void onCommand(WORD, WORD);
	void onNCPaint();
	void onTimer();

	void startConnect(const SocketAddress&);

	void serviceResolved(DNSServiceRef, std::string, std::string, uint16_t,
		const ServiceDiscovery::TXTRecord&);
	void serviceQueried(DNSServiceRef, std::string, uint16_t, uint16_t,
		const void*, uint32_t);

	const DeviceInfo _device;
	StreamSocket _socket;

	unsigned _milliseconds;
	volatile bool _connectStarted;
	uint16_t _port;
	DNSServiceRef _serviceRef;
};


//------------------------------------------------------------------------------


ConnectDialog::ConnectDialog(const DeviceInfo& device)
:
	_impl(new ConnectDialogImpl(device))
{
}


ConnectDialog::~ConnectDialog()
{
	delete _impl;
}


int ConnectDialog::doModal(HWND parentWindow)
{
	return _impl->doModal(DIALOG_CONNECT, parentWindow);
}


StreamSocket& ConnectDialog::socket() const
{
	return _impl->_socket;
}


//------------------------------------------------------------------------------


ConnectDialogImpl::ConnectDialogImpl(const DeviceInfo& device)
:
	_device(device),
	_milliseconds(0),
	_connectStarted(false),
	_serviceRef(0)
{
}


void ConnectDialogImpl::onInitialize()
{
	string_t name;
	Platform::Charset::fromUTF8(_device.name(), name);

	int returnCode;
	char_t formatString[256];
	char_t finalString[256];

	returnCode = GetDlgItemText(_dialogWindow, DIALOG_CONNECT_STRING,
		formatString, sizeof(formatString) / sizeof(char_t));
	assert(returnCode > 0 && returnCode < sizeof(formatString) / sizeof(char_t));

	returnCode = _stprintf_s(finalString, sizeof(finalString) / sizeof(char_t),
		formatString, name.c_str());
	assert(returnCode > 0 && returnCode < sizeof(finalString) / sizeof(char_t));

	returnCode = SetDlgItemText(_dialogWindow, DIALOG_CONNECT_STRING, finalString);
	assert(returnCode != 0);

	// start timer to facilitate showing dialog and updating progress bar
	if (SetTimer(_dialogWindow, (UINT_PTR) this, 250, NULL) < 1)
	{
		DWORD lastError = GetLastError();
		assert(lastError == 0);
	}

	if (!_device.isZeroConf())
	{
		startConnect(SocketAddress(_device.addr().first + ':' + _device.addr().second));
	}
	else
	{
		_serviceRef = ServiceDiscovery::resolveService(
			_device.addr().first, _device.addr().second, *this);
		ServiceDiscovery::start(_serviceRef);
	}
}


void ConnectDialogImpl::onCommand(WORD command, WORD control)
{
	if (control == IDCANCEL && command == BN_CLICKED)
	{
		if (_serviceRef)
			ServiceDiscovery::stop(_serviceRef);
		_socket.close();
		KillTimer(_dialogWindow, (UINT_PTR) this);
	}

	Dialog::onCommand(command, control);
}


void ConnectDialogImpl::onNCPaint()
{
	if (_milliseconds == 0)
	{
		// hide dialog initially
		ShowWindow(_dialogWindow, SW_HIDE);
	}
}


void ConnectDialogImpl::onTimer()
{
	if (!IsWindowVisible(_dialogWindow))
	{
		_milliseconds += 250;

		if (_milliseconds >= 1000)
		{
			// show dialog after 1 second
			ShowWindow(_dialogWindow, SW_SHOW);
		}
	}
	else
	{
		// update progress bar
		LRESULT progressLimit = SendDlgItemMessage(_dialogWindow,
			DIALOG_CONNECT_PROGRESS, PBM_GETRANGE, 0, 0);
		LRESULT progressValue = SendDlgItemMessage(_dialogWindow,
			DIALOG_CONNECT_PROGRESS, PBM_GETPOS, 0, 0);
		assert(progressLimit > 0);
		assert(progressValue >= 0);
		// increment progress value by 2% up to progress limit
		progressValue += (LRESULT) std::ceil((double) progressLimit * 0.02);
		progressValue %= progressLimit;
		SendDlgItemMessage(_dialogWindow, DIALOG_CONNECT_PROGRESS, PBM_SETPOS,
			(WPARAM) progressValue, 0);
	}

	if (_connectStarted)
	{
		// poll socket for writability
		if (_socket.poll(Timespan(), StreamSocket::SELECT_WRITE))
		{
			// set socket to blocking I/O since connection has been established
			_socket.setBlocking(true);

			KillTimer(_dialogWindow, (UINT_PTR) this);
			endDialog(0);
		}
	}
}


void ConnectDialogImpl::startConnect(const SocketAddress& addr)
{
	// start asynchronous socket connect
	_connectStarted = true;
	_socket.connectNB(addr);
}


void ConnectDialogImpl::serviceResolved(DNSServiceRef serviceRef,
	const std::string fullName, const std::string host, const uint16_t port,
	const ServiceDiscovery::TXTRecord& /*txtRecord*/)
{
	assert(_serviceRef == serviceRef);

	// stop service resolve activity
	ServiceDiscovery::stop(serviceRef);

	_port = port;

	// start service query activity
	_serviceRef = ServiceDiscovery::queryService(host, kDNSServiceType_A, *this);
	ServiceDiscovery::start(_serviceRef);
}


void ConnectDialogImpl::serviceQueried(DNSServiceRef serviceRef,
	const std::string fullName, const uint16_t /*rrtype*/, const uint16_t rdlen,
	const void* const rdata, const uint32_t /*ttl*/)
{
	assert(_serviceRef == serviceRef);

	// stop service query activity
	ServiceDiscovery::stop(serviceRef);

	_serviceRef = 0;

	startConnect(SocketAddress(IPAddress(rdata, rdlen), _port));
}
