/* Copyright (c) 2011  Eric Milles <eric.milles@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "Dialog.h"
#include <cassert>
#include <stdexcept>
#include <windowsx.h>


Poco::ThreadLocal<Dialog*> Dialog::_dialog;


INT_PTR CALLBACK Dialog::dialogProc(HWND dialogWindow, UINT message,
	WPARAM wParam, LPARAM lParam)
{
	// get dialog object pointer from window user data slot
	Dialog* dialog =
		reinterpret_cast<Dialog*>(GetWindowLongPtr(dialogWindow, GWLP_USERDATA));

	switch (message)
	{
	case WM_SETFONT:
		if (dialog == NULL)
		{
			// take dialog object pointer from thread-local storage
			dialog = _dialog.get();
			_dialog.get() = NULL;
			// put object pointer into window user data slot
			SetWindowLongPtr(dialogWindow, GWLP_USERDATA, (LONG_PTR) dialog);

			dialog->_dialogWindow = dialogWindow;
		}
		break;

	case WM_NCDESTROY:
		dialog->_dialogWindow = NULL;
		dialog->_dialogType = UNKNOWN;
		break;

	case WM_INITDIALOG:
		dialog->onInitialize();
		return TRUE;

	case WM_DESTROY:
		dialog->onFinalize();
		break;

	case WM_COMMAND:
		dialog->onCommand(HIWORD(wParam), LOWORD(wParam));
		break;

	case WM_CONTEXTMENU:
		{
			POINT point = { GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam) };
			dialog->onContextMenu(HWND(wParam), point);
		}
		break;

	case WM_NCPAINT:
		dialog->onNCPaint();
		break;

	case WM_TIMER:
		dialog->onTimer();
		break;

	case WM_COMPAREITEM:
		return dialog->onCompareItem(
			reinterpret_cast<LPCOMPAREITEMSTRUCT>(lParam));

	case WM_MEASUREITEM:
		return dialog->onMeasureItem(
			reinterpret_cast<LPMEASUREITEMSTRUCT>(lParam));

	case WM_DRAWITEM:
		return dialog->onDrawItem(
			reinterpret_cast<LPDRAWITEMSTRUCT>(lParam));
	}

	return FALSE;
}


//------------------------------------------------------------------------------


Dialog::Dialog()
:
	_dialogWindow(NULL),
	_dialogType(UNKNOWN)
{
}


Dialog::~Dialog()
{
	if (_dialogWindow != NULL)
	{
		endDialog(IDCANCEL);
	}
}


int Dialog::doModal(UINT templateId, HWND parentWindow)
{
	if (_dialogWindow == NULL)
	{
		_dialog.get() = this;
		_dialogType = MODAL;
		extern HINSTANCE dllInstance;
		return DialogBox(dllInstance, MAKEINTRESOURCE(templateId),
			parentWindow, Dialog::dialogProc);
	}
	else
	{
		throw std::runtime_error("_dialogWindow != NULL");
	}
}


void Dialog::doModeless(UINT templateId, HWND parentWindow)
{
	if (_dialogWindow == NULL)
	{
		_dialog.get() = this;
		_dialogType = MODELESS;
		extern HINSTANCE dllInstance;
		HWND dialog = CreateDialog(dllInstance, MAKEINTRESOURCE(templateId),
			parentWindow, Dialog::dialogProc);
		ShowWindow(dialog, SW_SHOW);
	}
	else
	{
		BringWindowToTop(_dialogWindow);
	}
}


void Dialog::endDialog(const int result)
{
	switch (_dialogType)
	{
	case MODAL:
		EndDialog(_dialogWindow, result);
		break;

	case MODELESS:
		DestroyWindow(_dialogWindow);
		break;
	}
}


void Dialog::onInitialize()
{
}


void Dialog::onFinalize()
{
}


void Dialog::onCommand(WORD command, WORD control)
{
	switch (control)
	{
	case IDCANCEL:
		if (command == BN_CLICKED)
		{
			endDialog(1);
		}
		break;

	case IDOK:
		if (command == BN_CLICKED)
		{
			endDialog(0);
		}
		break;
	}
}


void Dialog::onContextMenu(HWND /*window*/, POINT /*point*/)
{
}


void Dialog::onNCPaint()
{
}


void Dialog::onTimer()
{
}


int Dialog::onCompareItem(LPCOMPAREITEMSTRUCT /*compareItemInfo*/)
{
	return 0;
}


int Dialog::onMeasureItem(LPMEASUREITEMSTRUCT /*measureItemInfo*/)
{
	return 0;
}


int Dialog::onDrawItem(LPDRAWITEMSTRUCT /*drawItemInfo*/)
{
	return 0;
}
