/* Copyright (c) 2011  Eric Milles <eric.milles@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "MessageDialog.h"
#include "OptionsDialog.h"
#include "OptionsUtils.h"
#include "Output.h"
#include "Platform.h"
#include "Player.h"
#include "Plugin.h"
#include "Resources.h"
#include <cassert>
#include <cmath>
#include <functional>
#include <limits>
#include <memory>
#include <vector>
#include <wa_ipc.h>


Output Output::_this;
std::string Output::_iniFilePath;
std::auto_ptr<OptionsSubmenu> Output::_optionsSubmenu;
std::auto_ptr<OutputComponent> Output::_outputComponent;
Output::Progress Output::_bufferProgress;
Output::Progress Output::_outputProgress;
Output::Progress::second_type Output::_msecDivisor;


INT_PTR CALLBACK Output::optionsDialogProc(HWND dialogWindow, UINT message,
	WPARAM wParam, LPARAM /*lParam*/)
{
	static std::auto_ptr<OptionsDialog> optionsDialog;

	switch (message)
	{
	case WM_INITDIALOG:
		// center dialog window over parent window
		Platform::centerWindow(dialogWindow);

		optionsDialog.reset(new OptionsDialog());
		optionsDialog->setStatusCallback(
			// create callback that will update apply button
			std::tr1::bind(
				Platform::enableWindow,
				GetDlgItem(dialogWindow, IDOK),
				std::tr1::placeholders::_1));
		optionsDialog->doModeless(
			GetDlgItem(dialogWindow, DIALOG_OPTIONS_CHILD_FRAME));
		break;

	case WM_DESTROY:
		optionsDialog.reset();
		break;

	case WM_CLOSE:
		EndDialog(dialogWindow, 0);
		break;

	case WM_COMMAND:
		if (HIWORD(wParam) == BN_CLICKED)
		{
			switch (LOWORD(wParam))
			{
			case IDOK:
				optionsDialog->doApply();
				break;

			case IDCANCEL:
				EndDialog(dialogWindow, 0);
				break;
			}
		}
		break;
	}

	return 0;
}


void Output::init()
{
	extern std::auto_ptr<Player> player;

	// initialize options file path
	char* temp = (char*) SendMessage(player->window(), WM_WA_IPC, 0, IPC_GETINIFILE);
	if (temp != NULL)
	{
		_iniFilePath = temp;
	}
	else
	{
		extern HINSTANCE dllInstance;
		std::vector<char> pathBuffer(MAX_PATH);
		GetModuleFileNameA(dllInstance, &pathBuffer[0], pathBuffer.size());
		GetFullPathNameA(&pathBuffer[0], pathBuffer.size(), &pathBuffer[0], &temp);
		assert(temp != NULL);
		pathBuffer[temp - &pathBuffer[0]] = '\0';
		_iniFilePath = &pathBuffer[0];
		_iniFilePath.append("plugin.ini");
	}

	OptionsUtils::loadOptions(_iniFilePath);

	_optionsSubmenu.reset(new OptionsSubmenu(player->window()));

	_outputComponent.reset(new OutputComponent(*player));
	_outputComponent->setProgressCallback(&Output::bytesOutput);
}


void Output::quit()
{
	_outputComponent.reset();

	_optionsSubmenu.reset();

	OptionsUtils::saveOptions(_iniFilePath);
}


void Output::about(HWND parentWindow)
{
	MessageDialog(Plugin::aboutText().c_str(), 0).doModal(parentWindow);
}


void Output::options(HWND parentWindow)
{
	extern HINSTANCE dllInstance;
	INT_PTR result = DialogBox(dllInstance, MAKEINTRESOURCE(DIALOG_OPTIONS),
		parentWindow, &optionsDialogProc);
	if (result == -1)
	{
		DWORD lastError = GetLastError();
		assert(lastError == 0);
	}
}


int Output::open(const int samplesPerSecond, const int channelCount,
	const int bitsPerSample, const int /*bufferLengthMilliseconds*/,
	const int /*preBufferMilliseconds*/)
{
	try
	{
		const OutputFormat format(SampleRate(samplesPerSecond),
			SampleSize(bitsPerSample / 8), ChannelCount(channelCount));
		_outputComponent->open(format);

		_bufferProgress = _outputProgress = Progress(0, 0);
		_msecDivisor = static_cast<Progress::second_type>(
			samplesPerSecond * (bitsPerSample / 8) * channelCount) / 1000.0;

		const std::time_t latency = _outputComponent->latency();
		assert(latency <= static_cast<std::time_t>(std::numeric_limits<int>::max()));
		return static_cast<int>(latency);
	}
	catch (...)
	{
		return -1;
	}
}


void Output::close()
{
	_outputComponent->close();
}


void Output::reset(const int t)
{
	_outputComponent->reset();
	
	_bufferProgress = _outputProgress = Progress(t, 0);
}


int Output::write(char* const buffer, const int length)
{
	if (length >= 0 && _outputComponent->write(buffer, length))
	{
		updateProgress(_bufferProgress, length);
		return 0;
	}
	else
	{
		return 1;
	}
}


int Output::canWrite()
{
	const std::size_t size = _outputComponent->canWrite();
	assert(size <= static_cast<std::size_t>(std::numeric_limits<int>::max()));
	return static_cast<int>(size);
}


int Output::isPlaying()
{
	return (_outputComponent->buffered() > 0);
}


int Output::getOutputTime()
{
	// return millisecond counter from output progress
	return _outputProgress.first;
}


int Output::getWrittenTime()
{
	// return millisecond counter from buffer progress
	return _bufferProgress.first;
}


int Output::setPaused(const int state)
{
	return _outputComponent->setPaused(state != 0);
}


void Output::setBalance(const int balance)
{
	if (balance >= -128 && balance <= 128)
	{
		// convert balance from [-128,128] to [-1.0,1.0]
		_outputComponent->setBalance(static_cast<double>(balance) / 128.0);
	}
}


void Output::setVolume(const int volume)
{
	if (volume >= 0 && volume <= 255)
	{
		// convert volume from [0,255] to [0.0,1.0]
		_outputComponent->setVolume(static_cast<double>(volume) / 255.0);
	}
}


void Output::bytesOutput(const std::size_t bytes)
{
	updateProgress(_outputProgress, bytes);
}


void Output::updateProgress(Progress& progress, const std::size_t bytes)
{
	// increment byte counter
	progress.second += bytes;
	// increment millisecond counter with whole milliseconds
	progress.first += static_cast<Progress::first_type>(
		std::floor(progress.second / _msecDivisor));
	// update byte counter with remainder from bytes to milliseconds conversion
	progress.second = std::fmod(progress.second, _msecDivisor);
}
