/* Copyright (c) 2011  Eric Milles <eric.milles@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "OptionsUtils.h"
#include "Output.h"
#include "Plugin.h"
#include "XMPlayPlayer.h"
#include <cassert>
#include <memory>
#include <string>
#include <vector>


HINSTANCE dllInstance = NULL;
BYTE playerMajorVersion;
BYTE playerMinorVersion;
std::auto_ptr<Player> player;
static std::string iniFilePath;


/**
 * Main entry point for dynamic-link library (DLL).  This function is called
 * when the DLL is attached to or detached from a process or thread.
 *
 * @param instance
 * @param reason
 * @param reserved
 * @return <code>TRUE</code> on success, <code>FALSE</code> on failure
 */
BOOL WINAPI DllMain(HINSTANCE instance, DWORD reason, LPVOID /*reserved*/)
{
	switch (reason)
	{
	case DLL_PROCESS_ATTACH:
		assert(dllInstance == NULL);
		dllInstance = instance;
		break;

	case DLL_PROCESS_DETACH:
		OptionsUtils::saveOptions(iniFilePath);
		assert(dllInstance != NULL);
		dllInstance = NULL;
		break;
	}

	return TRUE;
}


/**
 * Main entry point for XMPlay output plug-in.  This function is called when the
 * player registers the plug-in.
 *
 * @return address of output structure to be used as interface between player
 *         and plug-in
 */
extern "C" __declspec(dllexport) XMPOUT* XMPOUT_GetInterface(DWORD face,
	InterfaceProc faceproc)
{
	if (face == 0)
	{
		// initialize output interface singleton
		static XMPOUT out = {
			0, // flags
			Plugin::name().c_str(),
			Output::optionsDialogProc,
			Output::getName,
			Output::getFlags,
			Output::open,
			Output::close,
			Output::reset,
			Output::setPaused,
			Output::setVolumeAndBalance,
			Output::canWrite,
			Output::write,
			Output::getBuffered,
			0, // getGeneralInfo callback
			0 // newTrack callback
		};

		if (player.get() == NULL)
		{
			XMPFUNC_MISC* const miscFunc =
				static_cast<XMPFUNC_MISC*>(faceproc(XMPFUNC_MISC_FACE));

			DWORD playerVersion = miscFunc->GetVersion();
			playerMajorVersion = HIBYTE(HIWORD(playerVersion));
			playerMinorVersion = LOBYTE(HIWORD(playerVersion));

			// initialize player interface singleton
			player.reset(new XMPlayPlayer(miscFunc));

			// initialize options file path
			std::vector<char> pathBuffer(MAX_PATH);
			GetModuleFileNameA(dllInstance, &pathBuffer[0], pathBuffer.size());
			char* temp;
			GetFullPathNameA(&pathBuffer[0], pathBuffer.size(), &pathBuffer[0], &temp);
			assert(temp != NULL);
			pathBuffer[temp - &pathBuffer[0]] = '\0';
			iniFilePath = &pathBuffer[0];
			iniFilePath.append("xmplay.ini");

			OptionsUtils::loadOptions(iniFilePath);
		}

		return &out;
	}
	else
	{
		return NULL;
	}
}
