/* Copyright (c) 2011  Eric Milles <eric.milles@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "OptionsDialog.h"
#include "Output.h"
#include "Platform.h"
#include "Player.h"
#include "Plugin.h"
#include "Resources.h"
#include <cassert>
#include <functional>
#include <limits>
#include <memory>


Output Output::_this;
std::auto_ptr<OutputComponent> Output::_outputComponent;
HANDLE Output::_outputEvent;


INT_PTR CALLBACK Output::optionsDialogProc(HWND dialogWindow, UINT message,
	WPARAM wParam, LPARAM /*lParam*/)
{
	static std::auto_ptr<OptionsDialog> optionsDialog;

	switch (message)
	{
	case WM_INITDIALOG:
		extern BYTE playerMajorVersion, playerMinorVersion;
		if (playerMajorVersion >= 0x03 && playerMinorVersion >= 0x05)
		{
			// v3.5 enlarged area for child frame
			const int dx = 0, dy = 24;

			// move apply button accordingly
			HWND applyButton =
				GetDlgItem(dialogWindow, DIALOG_OPTIONS_APPLY_BUTTON);
			RECT applyRect;
			GetWindowRect(applyButton, &applyRect);
			RECT dialogRect;
			GetWindowRect(dialogWindow, &dialogRect);
			// change from screen to dialog coordinates
			OffsetRect(&applyRect, -dialogRect.left, -dialogRect.top);
			// reposition apply button relative to its current position
			MoveWindow(applyButton,
				applyRect.left + dx, applyRect.top + dy,
				applyRect.right - applyRect.left,
				applyRect.bottom - applyRect.top,
				FALSE);
		}

		optionsDialog.reset(new OptionsDialog());
		optionsDialog->setStatusCallback(
			// create callback that will update apply button
			std::tr1::bind(
				Platform::enableWindow,
				GetDlgItem(dialogWindow, DIALOG_OPTIONS_APPLY_BUTTON),
				std::tr1::placeholders::_1));
		optionsDialog->doModeless(
			GetDlgItem(dialogWindow, DIALOG_OPTIONS_CHILD_FRAME));
		break;

	case WM_DESTROY:
		optionsDialog.reset();
		break;

	case WM_COMMAND:
		if (HIWORD(wParam) == BN_CLICKED)
		{
			switch (LOWORD(wParam))
			{
			case DIALOG_OPTIONS_APPLY_BUTTON:
				optionsDialog->doApply();
				break;
			}
		}
		break;
	}

	return 0;
}


const char* WINAPI Output::getName(DWORD face)
{
	return (face == 0 ? Plugin::name().c_str() : NULL);
}


DWORD WINAPI Output::getFlags(DWORD face)
{
	return (face == 0 ? XMPOUT_OUTPUT_VOLUME : -1);
}


BOOL WINAPI Output::open(DWORD face, XMPOUT_FORMAT* const form, HANDLE event)
{
	if (face == 0)
	{
		try
		{
			extern std::auto_ptr<Player> player;

			_outputComponent.reset(new OutputComponent(*player));
			_outputComponent->setProgressCallback(&Output::bytesOutput);

			const OutputFormat format(SampleRate(form->form.rate),
				SampleSize(form->form.res), ChannelCount(form->form.chan));
			_outputComponent->open(format);
			_outputEvent = event;

			return TRUE;
		}
		catch (...)
		{
			close();
		}
	}

	return FALSE;
}


void WINAPI Output::close()
{
	_outputComponent->close();
	_outputComponent.reset();
}


BOOL WINAPI Output::reset()
{
	_outputComponent->reset();
	return TRUE;
}


BOOL WINAPI Output::write(const void* const buffer, DWORD length)
{
	return _outputComponent->write(buffer, length);
}


DWORD WINAPI Output::canWrite()
{
	const std::size_t size = _outputComponent->canWrite();
	assert(size <= static_cast<std::size_t>(std::numeric_limits<DWORD>::max()));
	return static_cast<DWORD>(size);
}


DWORD WINAPI Output::getBuffered()
{
	const std::size_t size = _outputComponent->buffered();
	assert(size <= static_cast<std::size_t>(std::numeric_limits<DWORD>::max()));
	return static_cast<DWORD>(size);
}


BOOL WINAPI Output::setPaused(BOOL resume)
{
	_outputComponent->setPaused(resume == FALSE);
	return TRUE;
}


void WINAPI Output::setVolumeAndBalance(const float volume, const float balance)
{
	_outputComponent->setVolume(volume);
	_outputComponent->setBalance(balance);
}


void Output::bytesOutput(const std::size_t /*bytes*/)
{
	SetEvent(_outputEvent);
}
